"""
This module handles registering multimae models into timm.
"""
import logging
import torch
import numpy as np
from functools import partial
from huggingface_hub import hf_hub_download

from terratorch.datasets.utils import Modalities
from terratorch.models.backbones.multimae.multimae import MultiMAE, MultiViT
from terratorch.models.backbones.multimae.criterion import MaskedMSELoss, MaskedCrossEntropyLoss
from terratorch.models.backbones.multimae.input_adapters import PatchedInputAdapter, SemSegInputAdapter
from terratorch.models.backbones.multimae.output_adapters import SpatialOutputAdapter, ConvNeXtAdapter
from terratorch.registry import TERRATORCH_BACKBONE_REGISTRY, TERRATORCH_FULL_MODEL_REGISTRY

logger = logging.getLogger(__name__)

def _cfg(**kwargs) -> dict:
    return {
        "input_adapters": ['RGB'],
        "output_adapters": [],
        "dim_tokens": 768,
        "depth": 12,
        "num_heads": 12,
        "mlp_ratio": 4,
        "qkv_bias": True,
        "norm_layer": partial(torch.nn.LayerNorm, eps=1e-6),
        **kwargs,
    }

multimae_cfgs = {
    "multimae_small": _cfg(dim_tokens=384, depth=12, num_heads=6),
    "multimae_base": _cfg(),
    "multimae_large": _cfg(dim_tokens=1024, depth=24, num_heads=16),
}

pretrained_weights = {}

# TODO: Add pretrained models
# PRETRAINED_BANDS: list[HLSBands | int] = [
#     HLSBands.BLUE,
#     HLSBands.GREEN,
#     HLSBands.RED,
# ]


# TODO: make these user definable
DOMAIN_CONF = {
    Modalities.S1GRD: {
        "channels": 2,
        "stride_level": 1,
        "input_adapter": partial(PatchedInputAdapter, num_channels=2),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=2),
        "loss": MaskedMSELoss,
        "image_size": 224,
        "patch_size": 16,
    },
    Modalities.S1RTC: {
        "channels": 2,
        "stride_level": 1,
        "input_adapter": partial(PatchedInputAdapter, num_channels=2),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=2),
        "loss": MaskedMSELoss,
        "image_size": 224,
        "patch_size": 16,
    },
    Modalities.S2L1C: {
        "channels": 13,
        "stride_level": 1,
        "input_adapter": partial(PatchedInputAdapter, num_channels=13),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=13),
        "loss": MaskedMSELoss,
        "image_size": 224,
        "patch_size": 16,
    },
    Modalities.S2L2A: {
        "channels": 12,
        "stride_level": 1,
        "input_adapter": partial(PatchedInputAdapter, num_channels=12),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=12),
        "loss": MaskedMSELoss,
        "image_size": 224,
        "patch_size": 16,
    },
    Modalities.S2RGB: {
        "channels": 3,
        "stride_level": 1,
        "input_adapter": partial(PatchedInputAdapter, num_channels=3),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=3),
        "loss": MaskedMSELoss,
        "image_size": 224,
        "patch_size": 16,
    },
    Modalities.DEM: {
        "channels": 1,
        "stride_level": 1,
        "input_adapter": partial(PatchedInputAdapter, num_channels=1),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=1),
        "loss": MaskedMSELoss,
        "image_size": 224,
        "patch_size": 16,
    },
    Modalities.LULC: {
        "classes": 9,
        "stride_level": 1,
        "input_adapter": partial(SemSegInputAdapter, num_classes=9),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=9),  # Used in pretraining
        "loss": partial(MaskedCrossEntropyLoss, label_smoothing=0.0),
        "image_size": 224,
        "patch_size": 16,
    },
    'segmentation': {  # TODO: Test generalized semseg head from MultiMAE!
        "classes": 9,
        "stride_level": 1,
        "input_adapter": partial(SemSegInputAdapter, num_classes=9),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=9),  # Used in pretraining
        "loss": partial(MaskedCrossEntropyLoss, label_smoothing=0.0),
        "image_size": 224,
        "patch_size": 16,
    },
}


def _instantiate_input_adapter_from_dict(spec: dict) -> PatchedInputAdapter | SemSegInputAdapter:
    return spec["input_adapter"](
        stride_level=spec["stride_level"],
        patch_size_full=spec["patch_size"],
        image_size=spec["image_size"],
    )


def _parse_input_adapters(
    adapter_spec: list | dict[str, str | dict[str, int | str]],
) -> dict[str, PatchedInputAdapter | SemSegInputAdapter]:

    if isinstance(adapter_spec, list):
        # list to dict
        adapter_spec = {m: m for m in adapter_spec}
    if isinstance(adapter_spec, dict) and len(set(adapter_spec.keys())) != len(adapter_spec.keys()):
        msg = "Duplicate keys in input adapters"
        raise Exception(msg)
    input_adapters = {}

    for adapter_name, spec in adapter_spec.items():
        match spec:
            case str(spec):
                try:
                    spec = Modalities(spec.upper())
                except ValueError:
                    pass

                if spec in DOMAIN_CONF.keys():
                    input_adapters[adapter_name] = _instantiate_input_adapter_from_dict(DOMAIN_CONF[spec])
                else:
                    msg = f"Input Domain {adapter_name} does not exist. Choose one of {list(DOMAIN_CONF.keys())}"
                    raise ValueError(msg)
            case {"type": "PatchedInputAdapter", "num_channels": num_channels, **kwargs}:
                input_adapters[adapter_name] = PatchedInputAdapter(num_channels=num_channels, **kwargs)
            case {"type": "SemSegInputAdapter", "num_classes": num_classes,  **kwargs}:
                input_adapters[adapter_name] = SemSegInputAdapter(num_classes=num_classes, **kwargs)
            case _:
                msg = f"Invalid input adapter config for adapter {adapter_name}"
                raise ValueError(msg)
    return input_adapters


def _instantiate_output_adapter_from_dict(spec: dict, task: str, context_tasks: list[str], # num_channels: int
                                          ) -> SpatialOutputAdapter | ConvNeXtAdapter:
    return spec["output_adapter"](
        stride_level=spec["stride_level"],
        patch_size_full=spec["patch_size"],
        image_size=spec["image_size"],
        task=task,
        context_tasks=context_tasks,
        # num_channels=spec['channels'],  # TODO: Not passed in pretraining code
    )


def _instantiate_loss_from_dict(spec: dict) -> MaskedMSELoss | MaskedCrossEntropyLoss:
    return spec["loss"](
        patch_size=spec["patch_size"],
        stride=spec["stride_level"],
    )


def _parse_output_adapters(
    adapter_spec: list | dict[str, str | dict[str, int | str]],
) -> (dict[str, SpatialOutputAdapter | SpatialOutputAdapter], dict[str, MaskedMSELoss | MaskedCrossEntropyLoss]):

    if isinstance(adapter_spec, list):
        # list to dict
        adapter_spec = {m: m for m in adapter_spec}
    if isinstance(adapter_spec, dict) and len(set(adapter_spec.keys())) != len(adapter_spec.keys()):
        msg = "Duplicate keys in output adapters"
        raise Exception(msg)
    output_adapters = {}
    loss_functions = {}

    for adapter_name, spec in adapter_spec.items():
        match spec:
            case str(spec):
                try:
                    spec = Modalities(spec.upper())
                except ValueError:
                    pass

                if spec in DOMAIN_CONF.keys():
                    output_adapters[adapter_name] = _instantiate_output_adapter_from_dict(
                        DOMAIN_CONF[spec],
                        task=adapter_name,
                        context_tasks=list(adapter_spec.keys()),
                    )
                    loss_functions[adapter_name] = _instantiate_loss_from_dict(DOMAIN_CONF[spec])
                else:
                    msg = f"output Domain {adapter_name} does not exist. Choose one of {list(DOMAIN_CONF.keys())}"
                    raise ValueError(msg)
            case {"type": "SpatialOutputAdapter", "num_channels": num_channels, "patch_size": patch_size, **kwargs}:
                output_adapters[adapter_name] = SpatialOutputAdapter(
                    task=adapter_name,
                    context_tasks=list(adapter_spec.keys()),
                    patch_size_full=patch_size,
                    num_channels=num_channels,
                    **kwargs
                )
                loss_functions[adapter_name] = MaskedMSELoss(patch_size=patch_size)
            case {"type": "ConvNeXtAdapter", "num_classes": num_classes,  "patch_size": patch_size, **kwargs}:
                output_adapters[adapter_name] = ConvNeXtAdapter(
                    num_classes=num_classes,
                    patch_size=patch_size,
                    **kwargs
                )
                loss_functions[adapter_name] = MaskedCrossEntropyLoss(patch_size=patch_size)
            case _:
                msg = f"Invalid output adapter config for adapter {adapter_name}"
                raise ValueError(msg)
    return output_adapters, loss_functions


class PrepareMultimodalFeaturesForDecoder:
    def __init__(self, modalities: list[str], merging_method: str = 'concat'):
        self.modalities = modalities
        self.merging_method = merging_method

    def __call__(self, x: list[torch.Tensor]) -> list[torch.Tensor]:
        if len(x) == 2:
            # MultiMAE decoder was used. Return predictions of first modality.
            preds = list(x[0].values())
            assert len(preds) != 1, "Terratorch can only handle one output modality."
            return preds[0]

        for output_index in range(len(x)):
            x[output_index] = x[output_index].permute(0, 2, 1)
            x[output_index] = x[output_index][:, :, :-1]  # remove global token
            img_shape = int(np.sqrt(x[output_index].shape[-1] / len(self.modalities)))
            if self.merging_method == 'concat':
                x[output_index] = x[output_index].reshape(x[output_index].shape[0], -1, img_shape, img_shape)
            else:
                raise ValueError(f"Unsupported merging method {self.merging_method}")
                # TODO: Implement other methods, move to forward?

        return x


def _create_multimae(
    variant: str,
    pretrained: bool = False,
    ckpt_path: str = None,
    encoder_only: bool = True,
    merging_method: str = None,
    **kwargs,
):
    model_args = multimae_cfgs[variant].copy()
    model_args.update(kwargs)

    model_class = MultiViT if encoder_only else MultiMAE

    model = model_class(**model_args, merging_method=merging_method)

    if pretrained:
        if ckpt_path is not None:
            # Load model from checkpoint
            state_dict = torch.load(ckpt_path, map_location="cpu", weights_only=True)
        else:
            assert variant in pretrained_weights, (f"No pre-trained model found for variant {variant} "
                                                   f"(pretrained models: {pretrained_weights.keys()})")

            state_dict_file = hf_hub_download(repo_id=pretrained_weights[variant]['hf_hub_id'],
                                              filename=pretrained_weights[variant]['hf_hub_filename'])
            state_dict = torch.load(state_dict_file, map_location="cpu", weights_only=True)

        # TODO: add manual filtering of keys for strict weight loading
        loaded_keys = model.load_state_dict(state_dict, strict=False)
        if loaded_keys.missing_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")
        if loaded_keys.unexpected_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")
    elif ckpt_path is not None:
        logger.warning(f"ckpt_path is provided but pretrained is set to False, ignoring ckpt_path {ckpt_path}.")

    model.prepare_features_for_image_model = PrepareMultimodalFeaturesForDecoder(
        modalities=list(model.input_adapters.keys()),
        merging_method=merging_method
    )
    return model

@TERRATORCH_FULL_MODEL_REGISTRY.register
@TERRATORCH_BACKBONE_REGISTRY.register
def multimae_small(
    input_adapters: dict[str, str | dict[str, int | str]] | None = None,
    output_adapters: dict[str, str | dict[str, int | str]] | None = None,
    pretrained: bool = False,
    encoder_only: bool = True,
    **kwargs,
) -> torch.nn.Module:
    """MultiMAE small model."""

    if input_adapters is None:
        input_adapters = ['S1GRD', 'S1RTC', 'S2L1C', 'S2L2A', 'S2RGB', 'DEM', 'LULC']
        logger.warning(f'Using default adapters.')
    input_adapters = _parse_input_adapters(input_adapters)

    loss_functions = None
    if output_adapters is not None:
        output_adapters, loss_functions = _parse_output_adapters(output_adapters)
        encoder_only = False

    merging_method = None if output_adapters else kwargs.get('merging_method', 'concat')

    transformer = _create_multimae(
        "multimae_small",
        input_adapters=input_adapters,
        output_adapters=output_adapters,
        loss_functions=loss_functions,
        pretrained=pretrained,
        encoder_only=encoder_only,
        merging_method=merging_method,
        **kwargs,
    )
    return transformer

@TERRATORCH_FULL_MODEL_REGISTRY.register
@TERRATORCH_BACKBONE_REGISTRY.register
def multimae_base(
    input_adapters: dict[str, str | dict[str, int | str]] | None = None,
    output_adapters: dict[str, str | dict[str, int | str]] | None = None,
    pretrained: bool = False,
    encoder_only: bool = True,
    **kwargs,
) -> torch.nn.Module:
    """MultiMAE base model."""

    if input_adapters is None:
        input_adapters = ['S1GRD', 'S1RTC', 'S2L1C', 'S2L2A', 'S2RGB', 'DEM', 'LULC']
        logger.warning(f'Using default adapters.')
    input_adapters = _parse_input_adapters(input_adapters)

    loss_functions = None
    if output_adapters is not None:
        output_adapters, loss_functions = _parse_output_adapters(output_adapters)
        encoder_only = False

    merging_method = None if output_adapters else kwargs.get('merging_method', 'concat')

    transformer = _create_multimae(
        "multimae_base",
        input_adapters=input_adapters,
        output_adapters=output_adapters,
        loss_functions=loss_functions,
        pretrained=pretrained,
        encoder_only=encoder_only,
        merging_method=merging_method,
        **kwargs,
    )
    return transformer


@TERRATORCH_FULL_MODEL_REGISTRY.register
@TERRATORCH_BACKBONE_REGISTRY.register
def multimae_large(
    input_adapters: dict[str, str | dict[str, int | str]] | None = None,
    output_adapters: dict[str, str | dict[str, int | str]] | None = None,
    pretrained: bool = False,
    encoder_only: bool = True,
    **kwargs,
) -> torch.nn.Module:
    """MultiMAE large model."""

    if input_adapters is None:
        input_adapters = ['S1GRD', 'S1RTC', 'S2L1C', 'S2L2A', 'S2RGB', 'DEM', 'LULC']
        logger.warning(f'Using default adapters.')
    input_adapters = _parse_input_adapters(input_adapters)

    loss_functions = None
    if output_adapters is not None:
        output_adapters, loss_functions = _parse_output_adapters(output_adapters)
        encoder_only = False

    merging_method = None if output_adapters else kwargs.get('merging_method', 'concat')

    transformer = _create_multimae(
        "multimae_large",
        input_adapters=input_adapters,
        output_adapters=output_adapters,
        loss_functions=loss_functions,
        pretrained=pretrained,
        encoder_only=encoder_only,
        merging_method=merging_method,
        **kwargs,
    )
    return transformer



# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ---
#
# This project includes code adapted from the original work by EPFL and Apple Inc.,
# licensed under the Apache License, Version 2.0.
# Source: https://github.com/apple/ml-4m/

import math
import logging
import warnings
import torch
import torch.nn as nn
import torch.nn.functional as F
from functools import partial
from torch.amp import autocast
from einops import rearrange

# xFormers imports
try:
    from xformers.ops import memory_efficient_attention, unbind
    XFORMERS_AVAILABLE = True
except ImportError:
    logging.getLogger('terramind').debug("xFormers not available")
    XFORMERS_AVAILABLE = False


def pair(t):
    return t if isinstance(t, tuple) else (t, t)


def build_2d_sincos_posemb(h, w, embed_dim=1024, temperature=10000.):
    """Sine-cosine positional embeddings as used in MoCo-v3
    """
    grid_w = torch.arange(w, dtype=torch.float32)
    grid_h = torch.arange(h, dtype=torch.float32)
    grid_w, grid_h = torch.meshgrid(grid_w, grid_h, indexing='ij')
    assert embed_dim % 4 == 0, 'Embed dimension must be divisible by 4 for 2D sin-cos position embedding'
    pos_dim = embed_dim // 4
    omega = torch.arange(pos_dim, dtype=torch.float32) / pos_dim
    omega = 1. / (temperature ** omega)
    out_w = torch.einsum('m,d->md', [grid_w.flatten(), omega])
    out_h = torch.einsum('m,d->md', [grid_h.flatten(), omega])
    pos_emb = torch.cat([torch.sin(out_w), torch.cos(out_w), torch.sin(out_h), torch.cos(out_h)], dim=1)[None, :, :]
    pos_emb = rearrange(pos_emb, 'b (h w) d -> b d h w', h=h, w=w, d=embed_dim)
    return pos_emb


def _no_grad_trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1. + math.erf(x / math.sqrt(2.))) / 2.

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn("mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
                      "The distribution of values may be incorrect.",
                      stacklevel=2)

    with torch.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to
        # [2l-1, 2u-1].
        tensor.uniform_(2 * l - 1, 2 * u - 1)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        tensor.erfinv_()

        # Transform to proper mean, std
        tensor.mul_(std * math.sqrt(2.))
        tensor.add_(mean)

        # Clamp to ensure it's in the proper range
        tensor.clamp_(min=a, max=b)
        return tensor


def trunc_normal_(tensor, mean=0., std=1., a=-2., b=2.):
    # type: (Tensor, float, float, float, float) -> Tensor
    r"""Fills the input Tensor with values drawn from a truncated
    normal distribution. The values are effectively drawn from the
    normal distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`
    with values outside :math:`[a, b]` redrawn until they are within
    the bounds. The method used for generating the random values works
    best when :math:`a \leq \text{mean} \leq b`.
    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff value
        b: the maximum cutoff value
    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.trunc_normal_(w)
    """
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)


def drop_path(x, drop_prob: float = 0., training: bool = False):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).
    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.
    """
    if drop_prob == 0. or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).
    """

    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)

    def extra_repr(self) -> str:
        return 'p={}'.format(self.drop_prob)


class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        # x = self.drop(x)
        # commit this for the orignal BERT implement 
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads)

        if XFORMERS_AVAILABLE:
            q, k, v = unbind(qkv, 2) # Each is of shape B x N x num_heads x C // num_heads
            x = memory_efficient_attention(q, k, v)
            x = x.reshape([B, N, C])
        else:
            qkv = qkv.permute(2, 0, 3, 1, 4)
            q, k, v = qkv.unbind(0)   # make torchscript happy (cannot use tensor as tuple)

            attn = (q @ k.transpose(-2, -1)) * self.scale
            attn = attn.softmax(dim=-1)
            attn = self.attn_drop(attn)

            x = (attn @ v).transpose(1, 2).reshape(B, N, C)

        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class CrossAttention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim ** -0.5

        self.q = nn.Linear(dim, dim, bias=qkv_bias)
        self.kv = nn.Linear(dim, dim * 2, bias=qkv_bias)

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x, context):
        B, N, C = x.shape
        _, M, _ = context.shape

        q = self.q(x).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)
        kv = self.kv(context).reshape(B, M, 2, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        k, v = kv[0], kv[1]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, -1)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class Block(nn.Module):
    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, drop=0., attn_drop=0., 
                 drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.norm2 = norm_layer(dim)
        self.attn = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x, **kwargs):
        x = x + self.drop_path(self.attn(self.norm1(x)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class DecoderBlock(nn.Module):
    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, drop=0., attn_drop=0., 
                 drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.self_attn = Attention(dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop)
        self.cross_attn = CrossAttention(dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=drop)
        self.query_norm = norm_layer(dim)
        self.context_norm = norm_layer(dim)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x, context, **kwargs):
        x = x + self.drop_path(self.self_attn(self.norm1(x)))
        x = x + self.drop_path(self.cross_attn(self.query_norm(x), self.context_norm(context)))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class LayerNorm(nn.Module):
    r""" LayerNorm that supports two data formats: channels_last (default) or channels_first. 
    The ordering of the dimensions in the inputs. channels_last corresponds to inputs with 
    shape (batch_size, height, width, channels) while channels_first corresponds to inputs 
    with shape (batch_size, channels, height, width).

    From https://github.com/facebookresearch/ConvNeXt/blob/main/models/convnext.py
    """
    def __init__(self, normalized_shape, eps=1e-6, data_format="channels_last"):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError 
        self.normalized_shape = (normalized_shape, )
    
    def forward(self, x):
        if self.data_format == "channels_last":
            return F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        elif self.data_format == "channels_first":
            u = x.mean(1, keepdim=True)
            s = (x - u).pow(2).mean(1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            x = self.weight[:, None, None] * x + self.bias[:, None, None]
            return x

class ConvNeXtBlock(nn.Module):
    r""" ConvNeXt Block. There are two equivalent implementations:
    (1) DwConv -> LayerNorm (channels_first) -> 1x1 Conv -> GELU -> 1x1 Conv; all in (N, C, H, W)
    (2) DwConv -> Permute to (N, H, W, C); LayerNorm (channels_last) -> Linear -> GELU -> Linear; Permute back
    We use (2) as we find it slightly faster in PyTorch.

    From https://github.com/facebookresearch/ConvNeXt/blob/main/models/convnext.py
    
    Args:
        dim (int): Number of input channels.
        drop_path (float): Stochastic depth rate. Default: 0.0
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """
    def __init__(self, dim, drop_path=0., layer_scale_init_value=1e-6):
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, kernel_size=7, padding=3, groups=dim) # depthwise conv
        self.norm = nn.LayerNorm(dim, eps=1e-6)
        self.pwconv1 = nn.Linear(dim, 4 * dim) # pointwise/1x1 convs, implemented with linear layers
        self.act = nn.GELU()
        self.pwconv2 = nn.Linear(4 * dim, dim)
        self.gamma = nn.Parameter(layer_scale_init_value * torch.ones((dim)), 
                                    requires_grad=True) if layer_scale_init_value > 0 else None
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(self, x):
        input = x
        x = self.dwconv(x)
        x = x.permute(0, 2, 3, 1) # (N, C, H, W) -> (N, H, W, C)
        x = self.norm(x)
        x = self.pwconv1(x)
        x = self.act(x)
        x = self.pwconv2(x)
        if self.gamma is not None:
            x = self.gamma * x
        x = x.permute(0, 3, 1, 2) # (N, H, W, C) -> (N, C, H, W)

        x = input + self.drop_path(x)
        return x


class ViTEncoder(nn.Module):
    """Transformer to map images / feature maps to latent features.
    
    Args:
        in_channels: Number of input channels.
        patch_size: Patch size.
        resolution: Image resolution.
        dim_tokens: Transformer dimension.
        depth: Number of transformer layers.
        num_heads: Number of attention heads.
        mlp_ratio: MLP ratio.
        qkv_bias: If True, add bias to the qkv projection.
        drop_rate: Dropout rate.
        attn_drop_rate: Attention dropout rate.
        drop_path_rate: Stochastic depth rate.
        norm_layer: Normalization layer.
        sincos_pos_emb: If True, use sine-cosine positional embedding.
        learnable_pos_emb: If True, learn positional embedding.
        patch_proj: If True, project image patches to tokens.
          Consider disabling when encoding feature maps.
        post_mlp: If True, add MLP after transformer.
          See https://arxiv.org/abs/2110.04627.
        ckpt_path: Path to checkpoint to load.
    """
    def __init__(self, *, 
                 in_channels: int = 3, 
                 patch_size: int = 16,
                 resolution: int = 256,
                 dim_tokens: int = 768,
                 depth: int = 12,
                 num_heads: int = 12,
                 mlp_ratio: float = 4.0,
                 qkv_bias: bool = True,
                 drop_rate: float = 0.0,
                 attn_drop_rate: float = 0.0,
                 drop_path_rate: float = 0.0,
                 norm_layer: nn.Module = partial(nn.LayerNorm, eps=1e-6),
                 sincos_pos_emb: bool = True, 
                 learnable_pos_emb: bool = False, 
                 patch_proj: bool = True,
                 post_mlp: bool = False,
                 ckpt_path: str | None = None,
                 **ignore_kwargs):
        super().__init__()
        self.in_channels = in_channels
        self.P_H, self.P_W = pair(patch_size)
        self.H, self.W = pair(resolution)
        self.dim_tokens = dim_tokens
        self.patch_proj = patch_proj

        assert (self.H % self.P_H == 0) and (self.W % self.P_W == 0), f'Image sizes {self.H}x{self.W} must be divisible by patch sizes {self.P_H}x{self.P_W}'

        N_H = self.H // self.P_H
        N_W = self.W // self.P_W
        
        if sincos_pos_emb:
            self.pos_emb = build_2d_sincos_posemb(h=N_H, w=N_W, embed_dim=self.dim_tokens)
            self.pos_emb = nn.Parameter(self.pos_emb, requires_grad=learnable_pos_emb)
        else:
            self.pos_emb = nn.Parameter(torch.zeros(1, self.dim_tokens, N_H, N_W))
            trunc_normal_(self.pos_emb, std=0.02)

        # Image patches -> tokens projection
        if patch_proj:
            self.proj = nn.Conv2d(
                in_channels=self.in_channels, out_channels=self.dim_tokens,
                kernel_size=(self.P_H, self.P_W), stride=(self.P_H, self.P_W)
            )
        else:
            self.proj = nn.Conv2d(
                in_channels=self.in_channels, out_channels=self.dim_tokens,
                kernel_size=1, stride=1
            )

        # Transformer blocks
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.blocks = nn.Sequential(*[
            Block(dim=dim_tokens, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias,
                  drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i], norm_layer=norm_layer)
            for i in range(depth)
        ])

        if post_mlp:
            self.norm_mlp = norm_layer(dim_tokens)
            self.post_mlp = Mlp(dim_tokens, int(mlp_ratio*dim_tokens), act_layer=nn.Tanh)
        
        self.apply(self._init_weights)
        for name, m in self.named_modules():
            if isinstance(m, nn.Linear):
                if 'qkv' in name:
                    # treat the weights of Q, K, V separately
                    val = math.sqrt(6. / float(m.weight.shape[0] // 3 + m.weight.shape[1]))
                    nn.init.uniform_(m.weight, -val, val)
                elif 'kv' in name:
                    # treat the weights of K, V separately
                    val = math.sqrt(6. / float(m.weight.shape[0] // 2 + m.weight.shape[1]))
                    nn.init.uniform_(m.weight, -val, val)

            if isinstance(m, nn.Conv2d):
                if '.proj' in name:
                    # From MAE, initialize projection like nn.Linear (instead of nn.Conv2d)
                    w = m.weight.data
                    nn.init.xavier_uniform_(w.view([w.shape[0], -1]))

        if ckpt_path is not None:
            print(f'Loading checkpoint from {ckpt_path}')
            ckpt = torch.load(ckpt_path)
            ckpt['model']['pos_emb'] = rearrange(ckpt['model']['pos_embed'][:,1:], 'b (nh nw) d -> b d nh nw', nh=N_H, nw=N_W)
            ckpt['model']['proj.weight'] = ckpt['model']['patch_embed.proj.weight']
            ckpt['model']['proj.bias'] = ckpt['model']['patch_embed.proj.bias']
            msg = self.load_state_dict(ckpt['model'], strict=False)
            print(msg)

    def _init_weights(self, m: nn.Module) -> None:
        """Weight initialization"""
        if isinstance(m, nn.Linear):
            nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def get_num_layers(self) -> int:
        """Get number of transformer layers."""
        return len(self.blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """ViT encoder forward pass.
        
        Args:
            x: Input tensor of shape [B, C, H, W] or 
              [B, C, N_H, N_W] (patch projection disabled).

        Returns:
            Output tensor of shape [B, dim_tokens, N_H, N_W].
        """
        # Create patches [B, C, H, W] -> [B, (H*W), C]
        if self.patch_proj:
            B, C, H, W = x.shape
            assert (H % self.P_H == 0) and (W % self.P_W == 0), f'Image sizes {H}x{W} must be divisible by patch sizes {self.P_H}x{self.P_W}'
            N_H, N_W = H // self.P_H, W // self.P_W # Number of patches in height and width
        else:
            B, C, N_H, N_W = x.shape
        x = rearrange(self.proj(x), 'b d nh nw -> b (nh nw) d')

        if self.pos_emb is not None:
            # Create positional embedding
            x_pos_emb = F.interpolate(self.pos_emb, size=(N_H, N_W), mode='bicubic', align_corners=False)
            x_pos_emb = rearrange(x_pos_emb, 'b d nh nw -> b (nh nw) d')
            # Add positional embeddings to patches
            x = x + x_pos_emb

        # Transformer forward pass
        x = self.blocks(x)

        if hasattr(self, 'post_mlp'):
            with autocast('cuda', enabled=False):
                x = x.float() + self.post_mlp(self.norm_mlp(x.float()))

        # Reshape into 2D grid
        x = rearrange(x, 'b (nh nw) d -> b d nh nw', nh=N_H, nw=N_W)

        return x


class ViTDecoder(nn.Module):
    """Transformer to map latent features back to images / feature maps.
    
    Args:
        out_channels: Number of output channels.
        patch_size: Patch size.
        resolution: Image resolution.
        dim_tokens: Transformer dimension.
        depth: Number of transformer layers.
        num_heads: Number of attention heads.
        mlp_ratio: MLP ratio.
        qkv_bias: If True, add bias to the qkv projection.
        drop_rate: Dropout rate.
        attn_drop_rate: Attention dropout rate.
        drop_path_rate: Stochastic depth rate.
        norm_layer: Normalization layer.
        sincos_pos_emb: If True, use sine-cosine positional embedding.
        learnable_pos_emb: If True, learn positional embedding.
        patch_proj: If True, reproject tokens back to images.
          Consider disabling when encoding feature maps.
        post_mlp: If True, add MLP before transformer.
          See https://arxiv.org/abs/2110.04627.
        out_conv: If True, add two ConvNeXt blocks after transformer
          to deal with patch checkerboard artifacts.
    """
    def __init__(self, *, 
                 out_channels: int = 3, 
                 patch_size: int = 16,
                 resolution: int = 256,
                 dim_tokens: int = 768,
                 depth: int = 12,
                 num_heads: int = 12,
                 mlp_ratio: float = 4.0,
                 qkv_bias: bool = True,
                 drop_rate: float = 0.0,
                 attn_drop_rate: float = 0.0,
                 drop_path_rate: float = 0.0,
                 norm_layer: nn.Module = partial(nn.LayerNorm, eps=1e-6),
                 sincos_pos_emb: bool = True, 
                 learnable_pos_emb: bool = False,
                 patch_proj: bool = True,
                 post_mlp: bool = False,
                 out_conv: bool = False,
                 **ignore_kwargs):
        super().__init__()
        self.out_channels = out_channels
        self.P_H, self.P_W = pair(patch_size)
        self.H, self.W = pair(resolution)
        self.dim_tokens = dim_tokens
        self.patch_proj = patch_proj

        assert (self.H % self.P_H == 0) and (self.W % self.P_W == 0), f'Image sizes {self.H}x{self.W} must be divisible by patch sizes {self.P_H}x{self.P_W}'

        N_H = self.H // self.P_H
        N_W = self.W // self.P_W
        
        if sincos_pos_emb:
            self.pos_emb = build_2d_sincos_posemb(h=N_H, w=N_W, embed_dim=self.dim_tokens)
            self.pos_emb = nn.Parameter(self.pos_emb, requires_grad=learnable_pos_emb)
        else:
            self.pos_emb = nn.Parameter(torch.zeros(1, self.dim_tokens, N_H, N_W))
            trunc_normal_(self.pos_emb, std=0.02)

        # Transformer blocks
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.blocks = nn.Sequential(*[
            Block(dim=dim_tokens, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias,
                  drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i], norm_layer=norm_layer)
            for i in range(depth)
        ])

        # Tokens -> image output projection
        if post_mlp:
            self.norm_mlp = norm_layer(dim_tokens)
            self.post_mlp = Mlp(dim_tokens, int(mlp_ratio*dim_tokens), act_layer=nn.Tanh)
        if patch_proj:
            self.out_proj = nn.Linear(dim_tokens, self.out_channels * self.P_H * self.P_W)
        else:
            self.out_proj = nn.Linear(dim_tokens, self.out_channels)
        if out_conv:
            self.out_conv = nn.Sequential(ConvNeXtBlock(dim=self.out_channels), ConvNeXtBlock(dim=self.out_channels))
        
        self.apply(self._init_weights)
        for name, m in self.named_modules():
            if isinstance(m, nn.Linear):
                if 'qkv' in name:
                    # treat the weights of Q, K, V separately
                    val = math.sqrt(6. / float(m.weight.shape[0] // 3 + m.weight.shape[1]))
                    nn.init.uniform_(m.weight, -val, val)
                elif 'kv' in name:
                    # treat the weights of K, V separately
                    val = math.sqrt(6. / float(m.weight.shape[0] // 2 + m.weight.shape[1]))
                    nn.init.uniform_(m.weight, -val, val)

            if isinstance(m, nn.Conv2d):
                if '.proj' in name:
                    # From MAE, initialize projection like nn.Linear (instead of nn.Conv2d)
                    w = m.weight.data
                    nn.init.xavier_uniform_(w.view([w.shape[0], -1]))

    def _init_weights(self, m: nn.Module) -> None:
        """Weight initialization"""
        if isinstance(m, nn.Linear):
            nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def get_num_layers(self) -> int:
        """Get number of transformer layers."""
        return len(self.blocks)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """ViT decoder forward pass.
        
        Args:
            x: Input tensor of shape [B, dim_tokens, N_H, N_W].

        Returns:
            Output tensor of shape [B, C, H, W] or 
              [B, C, N_H, N_W] (patch projection disabled).
        """
        B, D, N_H, N_W = x.shape

        # Reshape into 1D
        x = rearrange(x, 'b d nh nw -> b (nh nw) d')

        if self.pos_emb is not None:
            # Create positional embedding
            x_pos_emb = F.interpolate(self.pos_emb, size=(N_H, N_W), mode='bicubic', align_corners=False)
            x_pos_emb = rearrange(x_pos_emb, 'b d nh nw -> b (nh nw) d')
            # Add positional embeddings to patches
            x = x + x_pos_emb

        # Transformer forward pass
        x = self.blocks(x)

        # Project each token to (C * P_H * P_W)
        if hasattr(self, 'post_mlp'):
            x = x + self.post_mlp(self.norm_mlp(x))
        x = self.out_proj(x)

        # Reshape sequence of patches into image or output features
        ph, pw = (self.P_H, self.P_W) if self.patch_proj else (1, 1)
        x = rearrange(
            x, 'b (nh nw) (c ph pw) -> b c (nh ph) (nw pw)',
            nh=N_H, nw=N_W, ph=ph, pw=pw, c=self.out_channels
        )

        # Optional conv layers to reduce patch artifacts
        if hasattr(self, 'out_conv'):
            x = self.out_conv(x)

        return x


# Encoder presets

def vit_s_enc(in_channels, 
              patch_size, 
              resolution, 
              drop_rate=0.0,
              attn_drop_rate=0.0,
              drop_path_rate=0.0,
              norm_layer=partial(nn.LayerNorm, eps=1e-6),
              sincos_pos_emb=True,
              learnable_pos_emb=False,
              patch_proj=True,
              post_mlp=False):
    model = ViTEncoder(
        in_channels=in_channels, 
        patch_size=patch_size,
        resolution=resolution,
        dim_tokens=512,
        depth=8,
        num_heads=8,
        mlp_ratio=4,
        qkv_bias=True,
        drop_rate=drop_rate,
        attn_drop_rate=attn_drop_rate,
        drop_path_rate=drop_path_rate,
        norm_layer=norm_layer,
        sincos_pos_emb=sincos_pos_emb,
        learnable_pos_emb=learnable_pos_emb,
        patch_proj=patch_proj,
        post_mlp=post_mlp,
    )
    return model

def vit_b_enc(in_channels, 
              patch_size, 
              resolution, 
              drop_rate=0.0,
              attn_drop_rate=0.0,
              drop_path_rate=0.0,
              norm_layer=partial(nn.LayerNorm, eps=1e-6),
              sincos_pos_emb=True,
              learnable_pos_emb=False,
              patch_proj=True,
              post_mlp=False,
              ckpt_path=None):
    model = ViTEncoder(
        in_channels=in_channels, 
        patch_size=patch_size,
        resolution=resolution,
        dim_tokens=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4,
        qkv_bias=True,
        drop_rate=drop_rate,
        attn_drop_rate=attn_drop_rate,
        drop_path_rate=drop_path_rate,
        norm_layer=norm_layer,
        sincos_pos_emb=sincos_pos_emb,
        learnable_pos_emb=learnable_pos_emb,
        patch_proj=patch_proj,
        post_mlp=post_mlp,
        ckpt_path=ckpt_path,
    )
    return model

def vit_l_enc(in_channels, 
              patch_size, 
              resolution, 
              drop_rate=0.0,
              attn_drop_rate=0.0,
              drop_path_rate=0.0,
              norm_layer=partial(nn.LayerNorm, eps=1e-6),
              sincos_pos_emb=True,
              learnable_pos_emb=False,
              patch_proj=True,
              post_mlp=False):
    model = ViTEncoder(
        in_channels=in_channels, 
        patch_size=patch_size,
        resolution=resolution,
        dim_tokens=1024,
        depth=24,
        num_heads=16,
        mlp_ratio=4,
        qkv_bias=True,
        drop_rate=drop_rate,
        attn_drop_rate=attn_drop_rate,
        drop_path_rate=drop_path_rate,
        norm_layer=norm_layer,
        sincos_pos_emb=sincos_pos_emb,
        learnable_pos_emb=learnable_pos_emb,
        patch_proj=patch_proj,
        post_mlp=post_mlp,
    )
    return model


# Decoder presets

def vit_s_dec(out_channels, 
              patch_size, 
              resolution, 
              drop_rate=0.0,
              attn_drop_rate=0.0,
              drop_path_rate=0.0,
              norm_layer=partial(nn.LayerNorm, eps=1e-6),
              sincos_pos_emb=True,
              learnable_pos_emb=False,
              patch_proj=True,
              post_mlp=False,
              out_conv=False):
    model = ViTDecoder(
        out_channels=out_channels, 
        patch_size=patch_size,
        resolution=resolution,
        dim_tokens=512,
        depth=8,
        num_heads=8,
        mlp_ratio=4,
        qkv_bias=True,
        drop_rate=drop_rate,
        attn_drop_rate=attn_drop_rate,
        drop_path_rate=drop_path_rate,
        norm_layer=norm_layer,
        sincos_pos_emb=sincos_pos_emb,
        learnable_pos_emb=learnable_pos_emb,
        patch_proj=patch_proj,
        post_mlp=post_mlp,
        out_conv=out_conv,
    )
    return model

def vit_b_dec(out_channels, 
              patch_size, 
              resolution, 
              drop_rate=0.0,
              attn_drop_rate=0.0,
              drop_path_rate=0.0,
              norm_layer=partial(nn.LayerNorm, eps=1e-6),
              sincos_pos_emb=True,
              learnable_pos_emb=False,
              patch_proj=True,
              post_mlp=False,
              out_conv=False):
    model = ViTDecoder(
        out_channels=out_channels, 
        patch_size=patch_size,
        resolution=resolution,
        dim_tokens=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4,
        qkv_bias=True,
        drop_rate=drop_rate,
        attn_drop_rate=attn_drop_rate,
        drop_path_rate=drop_path_rate,
        norm_layer=norm_layer,
        sincos_pos_emb=sincos_pos_emb,
        learnable_pos_emb=learnable_pos_emb,
        patch_proj=patch_proj,
        post_mlp=post_mlp,
        out_conv=out_conv,
    )
    return model

def vit_l_dec(out_channels, 
              patch_size, 
              resolution, 
              drop_rate=0.0,
              attn_drop_rate=0.0,
              drop_path_rate=0.0,
              norm_layer=partial(nn.LayerNorm, eps=1e-6),
              sincos_pos_emb=True,
              learnable_pos_emb=False,
              patch_proj=True,
              post_mlp=False,
              out_conv=False):
    model = ViTDecoder(
        out_channels=out_channels, 
        patch_size=patch_size,
        resolution=resolution,
        dim_tokens=1024,
        depth=24,
        num_heads=16,
        mlp_ratio=4,
        qkv_bias=True,
        drop_rate=drop_rate,
        attn_drop_rate=attn_drop_rate,
        drop_path_rate=drop_path_rate,
        norm_layer=norm_layer,
        sincos_pos_emb=sincos_pos_emb,
        learnable_pos_emb=learnable_pos_emb,
        patch_proj=patch_proj,
        post_mlp=post_mlp,
        out_conv=out_conv,
    )
    return model



from typing import Dict

import torch
from torch import nn
from torch.nn import Upsample

from terratorch.models.decoders.utils import ConvModule

def build_upsample_layer(cfg: Dict, **kwargs) -> nn.Module:
    """Build upsample layer.

    Args:
        cfg (dict): The upsample layer config, which should contain:

            - type (str): Layer type.
            - scale_factor (int): Upsample ratio, which is not applicable to
              deconv.
            - layer args: Args needed to instantiate a upsample layer.
        args (argument list): Arguments passed to the ``__init__``
            method of the corresponding conv layer.
        kwargs (keyword arguments): Keyword arguments passed to the
            ``__init__`` method of the corresponding conv layer.

    Returns:
        nn.Module: Created upsample layer.
    """
    # This config will allow the convolution to keep the input dimensions.
    standard_config = {"kernel_size": 3, "stride": 1, "padding": 1}

    upsample = Upsample
    
    upsample_layer = upsample(**cfg)
    conv_layer = ConvModule(
        **kwargs,
        **standard_config)

    layer = nn.Sequential(upsample_layer, conv_layer)

    return layer

class UpConvBlock(nn.Module):
    """Upsample convolution block in decoder for UNet.

    This upsample convolution block consists of one upsample module
    followed by one convolution block. The upsample module expands the
    high-level low-resolution feature map and the convolution block fuses
    the upsampled high-level low-resolution feature map and the low-level
    high-resolution feature map from encoder.

    Args:
        conv_block (nn.Sequential): Sequential of convolutional layers.
        in_channels (int): Number of input channels of the high-level
        skip_channels (int): Number of input channels of the low-level
        high-resolution feature map from encoder.
        out_channels (int): Number of output channels.
        num_convs (int): Number of convolutional layers in the conv_block.
            Default: 2.
        stride (int): Stride of convolutional layer in conv_block. Default: 1.
        dilation (int): Dilation rate of convolutional layer in conv_block.
            Default: 1.
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
        conv_cfg (dict | None): Config dict for convolution layer.
            Default: None.
        norm_cfg (dict | None): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict | None): Config dict for activation layer in ConvModule.
            Default: dict(type='ReLU').
        upsample_cfg (dict): The upsample config of the upsample module in
            decoder. Default: dict(type='InterpConv'). If the size of
            high-level feature map is the same as that of skip feature map
            (low-level feature map from encoder), it does not need upsample the
            high-level feature map and the upsample_cfg is None.
        dcn (bool): Use deformable convolution in convolutional layer or not.
            Default: None.
        plugins (dict): plugins for convolutional layers. Default: None.
    """

    def __init__(self,
                 conv_block,
                 in_channels,
                 skip_channels,
                 out_channels,
                 num_convs=2,
                 stride=1,
                 dilation=1,
                 with_cp=False,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(type='ReLU'),
                 upsample_cfg=None,
                 dcn=None,
                 plugins=None):
        super(UpConvBlock, self).__init__()
        assert dcn is None, 'Not implemented yet.'
        assert plugins is None, 'Not implemented yet.'

        self.conv_block = conv_block(
            in_channels=2 * skip_channels,
            out_channels=out_channels,
            num_convs=num_convs,
            stride=stride,
            dilation=dilation,
            with_cp=with_cp,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
            dcn=None,
            plugins=None)
        # TODO Upsampling as alternative for transposed convolution

        if upsample_cfg is not None:
            self.upsample = build_upsample_layer(
                cfg=upsample_cfg, in_channels=in_channels, out_channels=skip_channels) 
        else:
            self.upsample = ConvModule(
                in_channels,
                skip_channels,
                **self._default_conv_config,)
                #TODO add more options for configuring these layers
                #conv_cfg=conv_cfg,
                #norm_cfg=norm_cfg,
                #act_cfg=act_cfg)

    @property
    def _default_conv_config(self):

        return {"kernel_size": 4,
                "stride": 1,
                "padding": 1,
                "transpose": True,
                "scale_factor": 2}

    def forward(self, skip, x):
        """Forward function."""
        x = self.upsample(x)
        out = torch.cat([skip, x], dim=1)
        out = self.conv_block(out)
        return out
    
class BasicConvBlock(nn.Module):
    """Basic convolutional block for UNet.

    This module consists of several plain convolutional layers.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        num_convs (int): Number of convolutional layers. Default: 2.
        stride (int): Whether use stride convolution to downsample
            the input feature map. If stride=2, it only uses stride convolution
            in the first convolutional layer to downsample the input feature
            map. Options are 1 or 2. Default: 1.
        dilation (int): Whether use dilated convolution to expand the
            receptive field. Set dilation rate of each convolutional layer and
            the dilation rate of the first convolutional layer is always 1.
            Default: 1.
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
        conv_cfg (dict | None): Config dict for convolution layer.
            Default: None.
        norm_cfg (dict | None): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict | None): Config dict for activation layer in ConvModule.
            Default: dict(type='ReLU').
        dcn (bool): Use deformable convolution in convolutional layer or not.
            Default: None.
        plugins (dict): plugins for convolutional layers. Default: None.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 num_convs=2,
                 stride=1,
                 dilation=1,
                 with_cp=False,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(type='ReLU'),
                 dcn=None,
                 plugins=None):
        super(BasicConvBlock, self).__init__()
        assert dcn is None, 'Not implemented yet.'
        assert plugins is None, 'Not implemented yet.'

        self.with_cp = with_cp
        convs = []
        for i in range(num_convs):
            convs.append(
                ConvModule(
                    in_channels=in_channels if i == 0 else out_channels,
                    out_channels=out_channels,
                    kernel_size=3,
                    stride=stride if i == 0 else 1,
                    dilation=1 if i == 0 else dilation,
                    padding=1 if i == 0 else dilation,))
                    # TODO add more options for configuring these layers
                    #conv_cfg=conv_cfg,
                    #norm_cfg=norm_cfg,
                    #act_cfg=act_cfg))

        self.convs = nn.Sequential(*convs)

    def forward(self, x):
        """Forward function."""

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(self.convs, x)
        else:
            out = self.convs(x)
        return out


class DeconvModule(nn.Module):
    """Deconvolution upsample module in decoder for UNet (2X upsample).

    This module uses deconvolution to upsample feature map in the decoder
    of UNet.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
        norm_cfg (dict | None): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict | None): Config dict for activation layer in ConvModule.
            Default: dict(type='ReLU').
        kernel_size (int): Kernel size of the convolutional layer. Default: 4.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 with_cp=False,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(type='ReLU'),
                 *,
                 kernel_size=4,
                 scale_factor=2):
        super(DeconvModule, self).__init__()

        assert (kernel_size - scale_factor >= 0) and\
               (kernel_size - scale_factor) % 2 == 0,\
               f'kernel_size should be greater than or equal to scale_factor '\
               f'and (kernel_size - scale_factor) should be even numbers, '\
               f'while the kernel size is {kernel_size} and scale_factor is '\
               f'{scale_factor}.'

        stride = scale_factor
        padding = (kernel_size - scale_factor) // 2
        self.with_cp = with_cp
        deconv = nn.ConvTranspose2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding)

        norm_name, norm = build_norm_layer(norm_cfg, out_channels)
        activate = build_activation_layer(act_cfg)
        self.deconv_upsamping = nn.Sequential(deconv, norm, activate)

    def forward(self, x):
        """Forward function."""

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(self.deconv_upsamping, x)
        else:
            out = self.deconv_upsamping(x)
        return out


class InterpConv(nn.Module):
    """Interpolation upsample module in decoder for UNet.

    This module uses interpolation to upsample feature map in the decoder
    of UNet. It consists of one interpolation upsample layer and one
    convolutional layer. It can be one interpolation upsample layer followed
    by one convolutional layer (conv_first=False) or one convolutional layer
    followed by one interpolation upsample layer (conv_first=True).

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
        norm_cfg (dict | None): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict | None): Config dict for activation layer in ConvModule.
            Default: dict(type='ReLU').
        conv_cfg (dict | None): Config dict for convolution layer.
            Default: None.
        conv_first (bool): Whether convolutional layer or interpolation
            upsample layer first. Default: False. It means interpolation
            upsample layer followed by one convolutional layer.
        kernel_size (int): Kernel size of the convolutional layer. Default: 1.
        stride (int): Stride of the convolutional layer. Default: 1.
        padding (int): Padding of the convolutional layer. Default: 1.
        upsample_cfg (dict): Interpolation config of the upsample layer.
            Default: dict(
                scale_factor=2, mode='bilinear', align_corners=False).
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 with_cp=False,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(type='ReLU'),
                 *,
                 conv_cfg=None,
                 conv_first=False,
                 kernel_size=1,
                 stride=1,
                 padding=0,
                 upsample_cfg=dict(
                     scale_factor=2, mode='bilinear', align_corners=False)):
        super(InterpConv, self).__init__()

        self.with_cp = with_cp
        conv = ConvModule(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)
        upsample = Upsample(**upsample_cfg)
        if conv_first:
            self.interp_upsample = nn.Sequential(conv, upsample)
        else:
            self.interp_upsample = nn.Sequential(upsample, conv)

    def forward(self, x):
        """Forward function."""

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(self.interp_upsample, x)
        else:
            out = self.interp_upsample(x)
        return out






# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import os
import random
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Union

import albumentations as A
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio
import torch
from matplotlib.figure import Figure
from torch import Tensor

from terratorch.datasets.generic_multimodal_dataset import MultimodalToTensor
from terratorch.datasets.transforms import MultimodalTransforms
from terratorch.datasets.utils import default_transform
from torchgeo.datasets import BioMassters
from torchgeo.datasets.utils import percentile_normalization


class BioMasstersNonGeo(BioMassters):
    """[BioMassters Dataset](https://huggingface.co/datasets/ibm-nasa-geospatial/BioMassters) for Aboveground Biomass prediction.

    Dataset intended for Aboveground Biomass (AGB) prediction
    over Finnish forests based on Sentinel 1 and 2 data with
    corresponding target AGB mask values generated by Light Detection
    and Ranging (LiDAR).

    Dataset Format:

    * .tif files for Sentinel 1 and 2 data
    * .tif file for pixel wise AGB target mask
    * .csv files for metadata regarding features and targets

    Dataset Features:

    * 13,000 target AGB masks of size (256x256px)
    * 12 months of data per target mask
    * Sentinel 1 and Sentinel 2 data for each location
    * Sentinel 1 available for every month
    * Sentinel 2 available for almost every month
      (not available for every month due to ESA acquisition halt over the region
      during particular periods)

    If you use this dataset in your research, please cite the following paper:

    * https://nascetti-a.github.io/BioMasster/

    .. versionadded:: 0.5
    """

    S1_BAND_NAMES = ["VV_Asc", "VH_Asc", "VV_Desc", "VH_Desc", "RVI_Asc", "RVI_Desc"]
    S2_BAND_NAMES = [
        "BLUE",
        "GREEN",
        "RED",
        "RED_EDGE_1",
        "RED_EDGE_2",
        "RED_EDGE_3",
        "NIR_BROAD",
        "NIR_NARROW",
        "SWIR_1",
        "SWIR_2",
        "CLOUD_PROBABILITY",
    ]

    all_band_names = {
        "S1": S1_BAND_NAMES,
        "S2": S2_BAND_NAMES,
    }

    rgb_bands = {
        "S1": [],
        "S2": ["RED", "GREEN", "BLUE"],
    }

    valid_splits = ("train", "test")
    valid_sensors = ("S1", "S2")

    BAND_SETS = {"all": all_band_names, "rgb": rgb_bands}

    default_metadata_filename = "The_BioMassters_-_features_metadata.csv.csv"

    def __init__(
        self,
        root = "data",
        split: str = "train",
        bands: dict[str, Sequence[str]] | Sequence[str] = BAND_SETS["all"],
        transform: A.Compose | None = None,
        mask_mean: float | None = 63.4584,
        mask_std: float | None = 72.21242,
        sensors: Sequence[str] = ["S1", "S2"],
        as_time_series: bool = False,
        metadata_filename: str = default_metadata_filename,
        max_cloud_percentage: float | None = None,
        max_red_mean: float | None = None,
        include_corrupt: bool = True,
        subset: float = 1,
        seed: int = 42,
        use_four_frames: bool = False
    ) -> None:
        """Initialize a new instance of BioMassters dataset.

        If ``as_time_series=False`` (the default), each time step becomes its own
        sample with the target being shared across multiple samples.

        Args:
            root: root directory where dataset can be found
            split: train or test split
            sensors: which sensors to consider for the sample, Sentinel 1 and/or
                Sentinel 2 ('S1', 'S2')
            as_time_series: whether or not to return all available
                time-steps or just a single one for a given target location
            metadata_filename: metadata file to be used
            max_cloud_percentage: maximum allowed cloud percentage for images
            max_red_mean: maximum allowed red_mean value for images
            include_corrupt: whether to include images marked as corrupted

        Raises:
            AssertionError: if ``split`` or ``sensors`` is invalid
            DatasetNotFoundError: If dataset is not found.
        """
        self.root = root
        self.sensors = sensors
        self.bands = bands
        assert (
            split in self.valid_splits
        ), f"Please choose one of the valid splits: {self.valid_splits}."
        self.split = split

        assert set(sensors).issubset(
            set(self.valid_sensors)
        ), f"Please choose a subset of valid sensors: {self.valid_sensors}."

        if len(self.sensors) == 1:
            sens = self.sensors[0]
            self.band_indices = [
                self.all_band_names[sens].index(band) for band in self.bands[sens]
            ]
        else:
            self.band_indices = {
                sens: [self.all_band_names[sens].index(band) for band in self.bands[sens]]
                for sens in self.sensors
            }

        self.mask_mean = mask_mean
        self.mask_std = mask_std
        self.as_time_series = as_time_series
        self.metadata_filename = metadata_filename
        self.max_cloud_percentage = max_cloud_percentage
        self.max_red_mean = max_red_mean
        self.include_corrupt = include_corrupt
        self.subset = subset
        self.seed = seed
        self.use_four_frames = use_four_frames

        self._verify()

        # open metadata csv files
        self.df = pd.read_csv(os.path.join(self.root, self.metadata_filename))

        # Filter sensors
        self.df = self.df[self.df["satellite"].isin(self.sensors)]

        # Filter split
        self.df = self.df[self.df["split"] == self.split]

        # Optional filtering
        self._filter_and_select_data()

        # Optional subsampling
        self._random_subsample()

        # generate numerical month from filename since first month is September
        # and has numerical index of 0
        self.df["num_month"] = (
            self.df["filename"]
            .str.split("_", expand=True)[2]
            .str.split(".", expand=True)[0]
            .astype(int)
        )

        # Set dataframe index depending on the task for easier indexing
        if self.as_time_series:
            self.df["num_index"] = self.df.groupby(["chip_id"]).ngroup()
        else:
            filter_df = (
                self.df.groupby(["chip_id", "month"])["satellite"].count().reset_index()
            )
            filter_df = filter_df[
                filter_df["satellite"] == len(self.sensors)
            ].drop("satellite", axis=1)
            # Guarantee that each sample has corresponding number of images available
            self.df = self.df.merge(filter_df, on=["chip_id", "month"], how="inner")

            self.df["num_index"] = self.df.groupby(["chip_id", "month"]).ngroup()

        # Adjust transforms based on the number of sensors
        if len(self.sensors) == 1:
            self.transform = transform if transform else default_transform
        elif transform is None:
            self.transform = MultimodalToTensor(self.sensors)
        else:
            transform = {
                s: transform[s] if s in transform else default_transform
                for s in self.sensors
            }
            self.transform = MultimodalTransforms(transform, shared=False)

        if self.use_four_frames:
            self._select_4_frames()

    def __len__(self) -> int:
        return len(self.df["num_index"].unique())

    def _load_input(self, filenames: list[Path]) -> Tensor:
        """Load the input imagery at the index.

        Args:
            filenames: list of filenames corresponding to input

        Returns:
            input image
        """
        filepaths = [
            os.path.join(self.root, f"{self.split}_features", f) for f in filenames
        ]
        arr_list = [rasterio.open(fp).read() for fp in filepaths]

        if self.as_time_series:
            arr = np.stack(arr_list, axis=0) # (T, C, H, W)
        else:
            arr = np.concatenate(arr_list, axis=0)
        return arr.astype(np.int32)

    def _load_target(self, filename: Path) -> Tensor:
        """Load the target mask at the index.

        Args:
            filename: filename of target to index

        Returns:
            target mask
        """
        with rasterio.open(os.path.join(self.root, f"{self.split}_agbm", filename), "r") as src:
            arr: np.typing.NDArray[np.float64] = src.read()

        return arr

    def _compute_rvi(self, img: np.ndarray, linear: np.ndarray, sens: str) -> np.ndarray:
        """Compute the RVI indices for S1 data."""
        rvi_channels = []
        if self.as_time_series:
            if "RVI_Asc" in self.bands[sens]:
                try:
                    vv_asc_index = self.all_band_names["S1"].index("VV_Asc")
                    vh_asc_index = self.all_band_names["S1"].index("VH_Asc")
                except ValueError as e:
                    msg = f"RVI_Asc needs band: {e}"
                    raise ValueError(msg) from e

                VV = linear[:, vv_asc_index, :, :]
                VH = linear[:, vh_asc_index, :, :]
                rvi_asc = 4 * VH / (VV + VH + 1e-6)
                rvi_asc = np.expand_dims(rvi_asc, axis=1)
                rvi_channels.append(rvi_asc)
            if "RVI_Desc" in self.bands[sens]:
                try:
                    vv_desc_index = self.all_band_names["S1"].index("VV_Desc")
                    vh_desc_index = self.all_band_names["S1"].index("VH_Desc")
                except ValueError as e:
                    msg = f"RVI_Desc needs band: {e}"
                    raise ValueError(msg) from e

                VV_desc = linear[:, vv_desc_index, :, :]
                VH_desc = linear[:, vh_desc_index, :, :]
                rvi_desc = 4 * VH_desc / (VV_desc + VH_desc + 1e-6)
                rvi_desc = np.expand_dims(rvi_desc, axis=1)
                rvi_channels.append(rvi_desc)
            if rvi_channels:
                rvi_concat = np.concatenate(rvi_channels, axis=1)
                img = np.concatenate([img, rvi_concat], axis=1)
        else:
            if "RVI_Asc" in self.bands[sens]:
                if linear.shape[0] < 2:
                    msg = f"Not enough bands to calculate RVI_Asc. Available bands: {linear.shape[0]}"
                    raise ValueError(msg)
                VV = linear[0]
                VH = linear[1]
                rvi_asc = 4 * VH / (VV + VH + 1e-6)
                rvi_asc = np.expand_dims(rvi_asc, axis=0)
                rvi_channels.append(rvi_asc)
            if "RVI_Desc" in self.bands[sens]:
                if linear.shape[0] < 4:
                    msg = f"Not enough bands to calculate RVI_Desc. Available bands: {linear.shape[0]}"
                    raise ValueError(msg)
                VV_desc = linear[2]
                VH_desc = linear[3]
                rvi_desc = 4 * VH_desc / (VV_desc + VH_desc + 1e-6)
                rvi_desc = np.expand_dims(rvi_desc, axis=0) 
                rvi_channels.append(rvi_desc)
            if rvi_channels:
                rvi_concat = np.concatenate(rvi_channels, axis=0)
                img = np.concatenate([linear, rvi_concat], axis=0)
        return img

    def _select_4_frames(self):
        """Filter the dataset to select only 4 frames per sample."""

        if "cloud_percentage" in self.df.columns:
            self.df = self.df.sort_values(by=["chip_id", "cloud_percentage"])
        else:
            self.df = self.df.sort_values(by=["chip_id", "num_month"])

        self.df = (
            self.df.groupby("chip_id")
            .head(4)  # Select the first 4 frames per chip
            .reset_index(drop=True)
        )

    def _process_sensor_images(self, sens: str, sens_filepaths: list[str]) -> np.ndarray:
        """Process images for a given sensor."""
        img = self._load_input(sens_filepaths)
        if sens == "S1":
            img = img.astype(np.float32)
            linear = 10 ** (img / 10)
            img = self._compute_rvi(img, linear, sens)
        if self.as_time_series:
            img = img.transpose(0, 2, 3, 1)  # (T, H, W, C)
        else:
            img = img.transpose(1, 2, 0)  # (H, W, C)
        if len(self.sensors) == 1:
            img = img[..., self.band_indices]
        else:
            img = img[..., self.band_indices[sens]]
        return img

    def __getitem__(self, index: int) -> dict:
        sample_df = self.df[self.df["num_index"] == index].copy()
        # Sort by satellite and month
        sample_df.sort_values(
            by=["satellite", "num_month"], inplace=True, ascending=True
        )

        filepaths = sample_df["filename"].tolist()
        output = {}

        if len(self.sensors) == 1:
            sens = self.sensors[0]
            sens_filepaths = [fp for fp in filepaths if sens in fp]
            img = self._process_sensor_images(sens, sens_filepaths)
            output["image"] = img.astype(np.float32)
        else:
            for sens in self.sensors:
                sens_filepaths = [fp for fp in filepaths if sens in fp]
                img = self._process_sensor_images(sens, sens_filepaths)
                output[sens] = img.astype(np.float32)

        # Load target
        target_filename = sample_df["corresponding_agbm"].unique()[0]
        target = np.array(self._load_target(Path(target_filename)))
        target = target.transpose(1, 2, 0)
        output["mask"] = target
        if self.transform:
            if len(self.sensors) == 1:
                output = self.transform(**output)
            else:
                output = self.transform(output)
        output["mask"] = output["mask"].squeeze().float()
        return output

    def _filter_and_select_data(self):
        if (
            self.max_cloud_percentage is not None
            and "cloud_percentage" in self.df.columns
        ):
            self.df = self.df[self.df["cloud_percentage"] <= self.max_cloud_percentage]

        if self.max_red_mean is not None and "red_mean" in self.df.columns:
            self.df = self.df[self.df["red_mean"] <= self.max_red_mean]

        if not self.include_corrupt and "corrupt_values" in self.df.columns:
            self.df = self.df[self.df["corrupt_values"] is False]

    def _random_subsample(self):
        if self.split == "train" and self.subset < 1.0:
            num_samples = int(len(self.df["num_index"].unique()) * self.subset)
            if self.seed is not None:
                random.seed(self.seed)
            selected_indices = random.sample(
                list(self.df["num_index"].unique()), num_samples
            )
            self.df = self.df[self.df["num_index"].isin(selected_indices)]
            self.df.reset_index(drop=True, inplace=True)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample
        """
        # Determine if the sample contains multiple sensors or a single sensor
        if isinstance(sample["image"], dict):
            ncols = len(self.sensors) + 1
        else:
            ncols = 2  # One for the image and one for the mask

        showing_predictions = "prediction" in sample
        if showing_predictions:
            ncols += 1

        fig, axs = plt.subplots(1, ncols=ncols, figsize=(5 * ncols, 10))

        if isinstance(sample["image"], dict):
            # Multiple sensors case
            for idx, sens in enumerate(self.sensors):
                img = sample["image"][sens].numpy()
                if self.as_time_series:
                    # Plot last time step
                    img = img[:, -1, ...]
                if sens == "S2":
                    img = img[[2, 1, 0], ...].transpose(1, 2, 0)
                    img = percentile_normalization(img)
                else:
                    co_polarization = img[0]  # transmit == receive
                    cross_polarization = img[1]  # transmit != receive
                    ratio = co_polarization / (cross_polarization + 1e-6)

                    co_polarization = np.clip(co_polarization / 0.3, 0, 1)
                    cross_polarization = np.clip(cross_polarization / 0.05, 0, 1)
                    ratio = np.clip(ratio / 25, 0, 1)

                    img = np.stack(
                        (co_polarization, cross_polarization, ratio), axis=0
                    )
                    img = img.transpose(1, 2, 0)  # Convert to (H, W, 3)

                axs[idx].imshow(img)
                axs[idx].axis("off")
                if show_titles:
                    axs[idx].set_title(sens)
            mask_idx = len(self.sensors)
        else:
            # Single sensor case
            sens = self.sensors[0]
            img = sample["image"].numpy()
            if self.as_time_series:
                # Plot last time step
                img = img[:, -1, ...]
            if sens == "S2":
                img = img[[2, 1, 0], ...].transpose(1, 2, 0)
                img = percentile_normalization(img)
            else:
                co_polarization = img[0]  # transmit == receive
                cross_polarization = img[1]  # transmit != receive
                ratio = co_polarization / (cross_polarization + 1e-6)

                co_polarization = np.clip(co_polarization / 0.3, 0, 1)
                cross_polarization = np.clip(cross_polarization / 0.05, 0, 1)
                ratio = np.clip(ratio / 25, 0, 1)

                img = np.stack(
                    (co_polarization, cross_polarization, ratio), axis=0
                )
                img = img.transpose(1, 2, 0)  # Convert to (H, W, 3)

            axs[0].imshow(img)
            axs[0].axis("off")
            if show_titles:
                axs[0].set_title(sens)
            mask_idx = 1

        # Plot target mask
        if "mask" in sample:
            target = sample["mask"].squeeze()
            target_im = axs[mask_idx].imshow(target, cmap="YlGn")
            plt.colorbar(target_im, ax=axs[mask_idx], fraction=0.046, pad=0.04)
            axs[mask_idx].axis("off")
            if show_titles:
                axs[mask_idx].set_title("Target")

        # Plot prediction if available
        if showing_predictions:
            pred_idx = mask_idx + 1
            prediction = sample["prediction"].squeeze()
            pred_im = axs[pred_idx].imshow(prediction, cmap="YlGn")
            plt.colorbar(pred_im, ax=axs[pred_idx], fraction=0.046, pad=0.04)
            axs[pred_idx].axis("off")
            if show_titles:
                axs[pred_idx].set_title("Prediction")

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright contributors to the Terratorch project

from terratorch.datasets.fire_scars import FireScarsHLS, FireScarsNonGeo, FireScarsSegmentationMask
from terratorch.datasets.generic_pixel_wise_dataset import (
    GenericNonGeoPixelwiseRegressionDataset,
    GenericNonGeoSegmentationDataset,
)
from terratorch.datasets.generic_scalar_label_dataset import (
    GenericNonGeoClassificationDataset,
)
from terratorch.datasets.generic_multimodal_dataset import (
    GenericMultimodalDataset,
    GenericMultimodalSegmentationDataset,
    GenericMultimodalPixelwiseRegressionDataset,
    GenericMultimodalScalarDataset,
)
from terratorch.datasets.hls import HLSL30, HLSS30
from terratorch.datasets.m_bigearthnet import MBigEarthNonGeo
from terratorch.datasets.m_brick_kiln import MBrickKilnNonGeo

# geobench datasets segmentation
from terratorch.datasets.m_cashew_plantation import MBeninSmallHolderCashewsNonGeo
from terratorch.datasets.m_chesapeake_landcover import MChesapeakeLandcoverNonGeo

# geobench datasets classification
from terratorch.datasets.m_eurosat import MEuroSATNonGeo
from terratorch.datasets.m_forestnet import MForestNetNonGeo
from terratorch.datasets.m_neontree import MNeonTreeNonGeo
from terratorch.datasets.m_nz_cattle import MNzCattleNonGeo
from terratorch.datasets.m_pv4ger import MPv4gerNonGeo
from terratorch.datasets.m_pv4ger_seg import MPv4gerSegNonGeo
from terratorch.datasets.m_SA_crop_type import MSACropTypeNonGeo
from terratorch.datasets.m_so2sat import MSo2SatNonGeo
from terratorch.datasets.multi_temporal_crop_classification import MultiTemporalCropClassification
from terratorch.datasets.open_sentinel_map import OpenSentinelMap
from terratorch.datasets.pastis import PASTIS

# GenericNonGeoRegressionDataset,

from terratorch.datasets.sen1floods11 import Sen1Floods11NonGeo
from terratorch.datasets.utils import HLSBands, OpticalBands, SARBands

#from terratorch.datasets.sen1floods11 import Sen1Floods11NonGeo
from terratorch.datasets.sen4agrinet import Sen4AgriNet

from terratorch.datasets.burn_intensity import BurnIntensityNonGeo
from terratorch.datasets.carbonflux import CarbonFluxNonGeo
from terratorch.datasets.landslide4sense import Landslide4SenseNonGeo
from terratorch.datasets.forestnet import ForestNetNonGeo
from terratorch.datasets.biomassters import BioMasstersNonGeo

# TorchGeo RasterDatasets
from terratorch.datasets.wsf import WSF2019, WSFEvolution


# miscellaneous datasets
from terratorch.datasets.openearthmap import OpenEarthMapNonGeo

# Generic Classification Dataset
from terratorch.datasets.sen4map import Sen4MapDatasetMonthlyComposites


__all__ = (
    "GenericNonGeoSegmentationDataset",
    "GenericNonGeoPixelwiseRegressionDataset",
    "GenericNonGeoClassificationDataset",
    # "GenericNonGeoRegressionDataset",
    "GenericMultimodalDataset",
    "GenericMultimodalSegmentationDataset",
    "GenericMultimodalPixelwiseRegressionDataset",
    "GenericMultimodalScalarDataset",
    "GenericNonGeoRegressionDataset",
    "BurnIntensityNonGeo",
    "CarbonFluxNonGeo",
    "Landslide4SenseNonGeo",
    "BioMasstersNonGeo",
    "ForestNetNonGeo",
    "FireScarsNonGeo",
    "FireScarsHLS",
    "FireScarsSegmentationMask",
    "Sen1Floods11NonGeo",
    "MultiTemporalCropClassification",
    "Sen4MapDatasetMonthlyComposites",
    "HLSBands",
    "MEuroSATNonGeo",
    "MBigEarthNonGeo",
    "MBrickKilnNonGeo",
    "MForestNetNonGeo",
    "MSo2SatNonGeo",
    "MPv4gerNonGeo",
    "MBeninSmallHolderCashewsNonGeo",
    "MNzCattleNonGeo",
    "MChesapeakeLandcoverNonGeo",
    "MPv4gerSegNonGeo",
    "MSACropTypeNonGeo",
    "MNeonTreeNonGeo",
    "OpenSentinelMap",
    "PASTIS",
    "Sen4AgriNet",
    "WSF2019",
    "WSFEvolution",
    "HLSL30",
    "HLSS30",
    "OpticalBands",
    "SARBands",
    "OpenEarthMapNonGeo"
)



# Copyright (c) EPFL VILAB.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# --------------------------------------------------------
# Based on timm, DeiT, DINO, MoCo-v3, BEiT, MAE-priv and MAE code bases
# https://github.com/rwightman/pytorch-image-models/tree/master/timm
# https://github.com/facebookresearch/deit
# https://github.com/facebookresearch/dino
# https://github.com/facebookresearch/moco-v3
# https://github.com/microsoft/unilm/tree/master/beit
# https://github.com/BUPT-PRIV/MAE-priv
# https://github.com/facebookresearch/mae
# --------------------------------------------------------

import itertools
import math
from collections import OrderedDict
from functools import partial

import torch
from einops import repeat
from torch import nn
from torch.distributions.dirichlet import Dirichlet
import torch.nn.functional as F
from .multimae_utils import Block, trunc_normal_


class MultiMAE(nn.Module):
    """MultiMAE: Multi-task Multi-modal Masked Autoencoder
    This module performs masking in its forward pass.
    The MultiViT module defined below inherits from this module and performs a regular forward pass,
    and should be used instead for downstream tasks


    :param input_adapters: Dictionary of task -> input adapters
    :param output_adapters: Optional dictionary of task -> output adapters

    :param num_global_tokens: Number of additional global tokens to add (like cls tokens), default is 1
    :param dim_tokens: Dimension of encoder tokens
    :param depth: Depth of encoder
    :param num_heads: Number of attention heads
    :param mlp_ratio: MLP hidden dim ratio
    :param qkv_bias: Set to False to disable bias
    :param drop_rate: Dropout after MLPs and Attention
    :param attn_drop_rate: Attention matrix drop rate
    :param drop_path_rate: DropPath drop rate
    :param norm_layer: Type of normalization layer
    :param fp32_output_adapters: List of task identifiers to force output adapters to
    run with mixed precision turned off for stability reasons.
    """

    default_norm_layer = partial(nn.LayerNorm, eps=1e-6)

    def __init__(
        self,
        input_adapters: dict[str, nn.Module],
        output_adapters: dict[str, nn.Module] | None,
        loss_functions: dict[str, nn.Module] | None,
        num_global_tokens: int = 1,
        dim_tokens: int = 768,
        depth: int = 12,
        num_heads: int = 12,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        drop_rate: float = 0.0,
        attn_drop_rate: float = 0.0,
        drop_path_rate: float = 0.0,
        norm_layer: nn.Module = default_norm_layer,
        fp32_output_adapters: list[str] | None = None,
        num_input_tokens: int = 128,
        merge_method: str = None,
        **kwargs,
    ):
        super().__init__()

        # Initialize input and output adapters
        for adapter in input_adapters.values():
            adapter.init(dim_tokens=dim_tokens)
        self.input_adapters = nn.ModuleDict(input_adapters)
        if output_adapters is not None:
            for adapter in output_adapters.values():
                adapter.init(dim_tokens_enc=dim_tokens)
            self.output_adapters = nn.ModuleDict(output_adapters)
        else:
            self.output_adapters = None
        self.fp32_output_adapters = fp32_output_adapters or []
        self.loss_functions = loss_functions
        self.num_input_tokens = num_input_tokens

        # Additional learnable tokens that can be used by encoder to process/store global information
        self.num_global_tokens = num_global_tokens
        self.global_tokens = nn.Parameter(torch.zeros(1, num_global_tokens, dim_tokens))
        trunc_normal_(self.global_tokens, std=0.02)

        if merge_method == 'concat':  # TODO: Move prepare/concat to this model forward?
            embed_factor = len(input_adapters)
        else:
            embed_factor = 1
        self.merge_method = merge_method
        assert merge_method in ['mean', 'max', 'concat', None], "merge_method must be one of mean, max, concat, None."
        self.out_channels = [int(dim_tokens) * embed_factor] * depth

        # Transformer encoder
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule

        self.layers: nn.ModuleList = nn.ModuleList(
            Block(
                dim=dim_tokens,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr[i],
                norm_layer=norm_layer,
            )
            for i in range(depth)
        )

        self.apply(self._init_weights)
        for name, m in self.named_modules():
            if isinstance(m, nn.Linear):
                if "qkv" in name:
                    # treat the weights of Q, K, V separately
                    val = math.sqrt(6.0 / float(m.weight.shape[0] // 3 + m.weight.shape[1]))
                    nn.init.uniform_(m.weight, -val, val)
                elif "kv" in name:
                    # treat the weights of K, V separately
                    val = math.sqrt(6.0 / float(m.weight.shape[0] // 2 + m.weight.shape[1]))
                    nn.init.uniform_(m.weight, -val, val)

            if isinstance(m, nn.Conv2d):
                if ".proj" in name:
                    # From MAE, initialize projection like nn.Linear (instead of nn.Conv2d)
                    w = m.weight.data
                    nn.init.xavier_uniform_(w.view([w.shape[0], -1]))

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            nn.init.xavier_uniform_(m.weight)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def get_num_layers(self):
        return len(self.layers)

    @torch.jit.ignore
    def no_weight_decay(self):
        no_wd_set = {"global_tokens"}

        for task, adapter in self.input_adapters.items():
            if hasattr(adapter, "no_weight_decay"):
                to_skip = adapter.no_weight_decay()
                to_skip = {f"input_adapters.{task}.{name}" for name in to_skip}
                no_wd_set = no_wd_set | to_skip

        for task, adapter in self.output_adapters.items():
            if hasattr(adapter, "no_weight_decay"):
                to_skip = adapter.no_weight_decay()
                to_skip = {f"output_adapters.{task}.{name}" for name in to_skip}
                no_wd_set = no_wd_set | to_skip

        return no_wd_set

    def sample_alphas(self, B: int, n_tasks: int, alphas: float = 1.0, eps: float = 1e-5):
        """
        Sample alphas for Dirichlet sampling such that tasks are first uniformly chosen and then Dirichlet sampling
        is performed over the chosen ones.

        :param B: Batch size
        :param n_tasks: Number of input tasks
        :param alphas: Float or list to multiply task choices {0,1} by
        :param eps: Small constant since Dirichlet alphas need to be positive
        """
        valid_task_choices = torch.Tensor([list(i) for i in itertools.product([0, 1], repeat=n_tasks)][1:])
        rand_per_sample_choice = torch.randint(0, len(valid_task_choices), (B,))
        alphas_tensor = torch.index_select(valid_task_choices, 0, rand_per_sample_choice)
        alphas_tensor = alphas_tensor * torch.tensor(alphas) + eps
        return alphas_tensor

    def generate_random_masks(
        self,
        input_tokens: dict[str, torch.Tensor],
        num_input_tokens: int,
        alphas: float | list[float] = 1.0,
        sample_tasks_uniformly: bool = False,
    ):
        """
        Sample a total of num_input_tokens from different tasks using Dirichlet sampling.

        :param input_tokens: Dictionary of tensors to sample num_input_tokens from
        :param num_input_tokens: Number of tokens to select
        :param alphas: Dirichlet distribution parameter alpha. Lower alpha = harder,
            less uniform sampling. Can be float or list of floats.
        :param sample_tasks_uniformly: Set to True to first sample 1-n_tasks uniformly at random
            for each sample in the batch. Dirichlet sampling is then done over selected subsets.
        """
        B = list(input_tokens.values())[0].shape[0]
        device = next(iter(input_tokens.values())).device

        alphas = [alphas] * len(input_tokens) if isinstance(alphas, float) else alphas
        if sample_tasks_uniformly:
            alphas = self.sample_alphas(B, len(input_tokens), alphas=alphas)
            task_sampling_dist = Dirichlet(alphas).sample().to(device)
        else:
            task_sampling_dist = Dirichlet(torch.Tensor(alphas)).sample((B,)).to(device)

        samples_per_task = (task_sampling_dist * num_input_tokens).round().long()

        task_masks = []
        num_tokens_per_task = [task_tokens.shape[1] for task_tokens in input_tokens.values()]
        for i, num_tokens in enumerate(num_tokens_per_task):
            # Use noise to shuffle arange
            noise = torch.rand(B, num_tokens, device=device)  # noise in [0, 1]
            ids_arange_shuffle = torch.argsort(noise, dim=1)  # ascend: small is keep, large is remove
            mask = torch.arange(num_tokens, device=device).unsqueeze(0).expand(B, -1)
            mask = torch.gather(mask, dim=1, index=ids_arange_shuffle)
            # 0 is keep (unmasked), 1 is remove (masked)
            mask = torch.where(mask < samples_per_task[:, i].unsqueeze(1), 0, 1)
            task_masks.append(mask)

        mask_all = torch.cat(task_masks, dim=1)
        ids_shuffle = torch.argsort(mask_all + torch.rand_like(mask_all.float()), dim=1)
        ids_restore = torch.argsort(ids_shuffle, dim=1)
        ids_keep = ids_shuffle[:, :num_input_tokens]

        # Update binary mask to adjust for task rounding
        mask_all = torch.ones_like(mask_all)
        mask_all[:, :num_input_tokens] = 0
        # Unshuffle to get the binary mask
        mask_all = torch.gather(mask_all, dim=1, index=ids_restore)
        # Split to get task masks
        task_masks = torch.split(mask_all, num_tokens_per_task, dim=1)
        # Convert to dict
        task_masks = dict(zip(input_tokens.keys(), task_masks, strict=False))

        return task_masks, ids_keep, ids_restore

    @staticmethod
    def make_mask(
        N_H,
        N_W,
        xy_idxs,
        full_tasks=[],
        indicate_visible=True,
        flatten=True,
        device="cuda",
    ):
        """
        Creates masks for each task, given lists of un-masked x,y coordinates.
        """
        xy_idxs = {k: torch.LongTensor(v) for k, v in xy_idxs.items()}

        task_masks = {k: torch.ones(N_H, N_W).to(device) for k in xy_idxs.keys()}

        for k in xy_idxs.keys():
            if len(xy_idxs[k]) > 0:
                task_masks[k][xy_idxs[k][:, 1], xy_idxs[k][:, 0]] = 0

        for task in full_tasks:
            task_masks[task][:] = 0

        if not indicate_visible:
            task_masks = {k: 1 - v for k, v in task_masks.items()}

        if flatten:
            task_masks = {k: v.flatten().unsqueeze(0) for k, v in task_masks.items()}

        return task_masks

    def generate_input_info(self, input_task_tokens, image_size):
        input_info = OrderedDict()
        i = 0
        input_info["tasks"] = {}
        for domain, tensor in input_task_tokens.items():
            num_tokens = tensor.shape[1]
            d = {
                "num_tokens": num_tokens,
                "has_2d_posemb": True,  # TODO: Modify when adding non-2D tasks
                "start_idx": i,
                "end_idx": i + num_tokens,
            }
            i += num_tokens
            input_info["tasks"][domain] = d

        input_info["image_size"] = image_size
        input_info["num_task_tokens"] = i
        input_info["num_global_tokens"] = self.num_global_tokens

        return input_info

    def forward(
        self,
        x: dict[str, torch.Tensor] | torch.Tensor,
        mask_inputs: bool = True,
        task_masks: dict[str, torch.Tensor] = None,
        num_input_tokens: int | None = None,
        alphas: float | list[float] = 0.5,
        sample_tasks_uniformly: bool = False,
    ):
        """
        Forward pass through input adapters, transformer encoder and output adapters.
        If specified, will randomly drop input tokens.

        :param x: Input tensor or dictionary of tensors
        :param mask_inputs: Set to True to enable random masking of input patches
        :param task_masks: Optional dictionary of task->mask pairs.
        :param num_input_tokens: Number of tokens to randomly select for encoder.
            Only used if mask_inputs is True.
        :param alphas: Dirichlet distribution parameter alpha for task sampling.
            Higher alpha = harder, less uniform sampling. Can be float or list of floats.
        :param sample_tasks_uniformly: Set to True if tasks should be uniformly presampled,
            before Dirichlet sampling decides share of masked tokens between them.
        """

        ## Processing input modalities
        # If input x is a Tensor, assume it's RGB
        x = {"rgb": x} if isinstance(x, torch.Tensor) else x

        # Need image size for tokens->image reconstruction
        # Assuming same image size for all modalities
        shape = list(x.values())[0].shape
        B = shape[0]
        H, W = shape[-2:]

        # Encode selected inputs to tokens
        input_task_tokens = {
            domain: self.input_adapters[domain](tensor)
            for domain, tensor in x.items()
            if domain in self.input_adapters
        }

        input_info = self.generate_input_info(input_task_tokens=input_task_tokens, image_size=(H, W))

        # Select random subset of tokens from the chosen input tasks and concatenate them
        if mask_inputs:
            num_input_tokens = num_input_tokens or self.num_input_tokens
        else:
            num_input_tokens = sum([tensor.shape[1] for tensor in input_task_tokens.values()])

        # Generating masks
        if task_masks is None:
            task_masks, ids_keep, ids_restore = self.generate_random_masks(
                input_task_tokens, num_input_tokens, alphas=alphas, sample_tasks_uniformly=sample_tasks_uniformly
            )
        else:
            mask_all = torch.cat([task_masks[task] for task in input_task_tokens.keys()], dim=1)
            ids_shuffle = torch.argsort(mask_all, dim=1)
            ids_restore = torch.argsort(ids_shuffle, dim=1)
            ids_keep = ids_shuffle[:, : (mask_all == 0).sum()]

        input_tokens = torch.cat([task_tokens for task_tokens in input_task_tokens.values()], dim=1)

        # Apply mask
        input_tokens = torch.gather(input_tokens, dim=1,
                                    index=ids_keep.unsqueeze(-1).repeat(1, 1, input_tokens.shape[2]))

        # Add global tokens to input tokens
        global_tokens = repeat(self.global_tokens, "() n d -> b n d", b=B)
        input_tokens = torch.cat([input_tokens, global_tokens], dim=1)

        # Transformer forward pass
        outputs = []
        encoder_tokens = input_tokens
        for layer in self.layers:
            encoder_tokens = layer(encoder_tokens)
            outputs.append(encoder_tokens)

        # Output decoders
        if self.output_adapters is None:
            return outputs, task_masks

        # TODO: Add target masking
        # Decode tokens for each task using task-specific output adapters
        preds = {
            domain: self.output_adapters[domain](
                encoder_tokens=encoder_tokens,
                input_info=input_info,
                ids_keep=ids_keep,
                ids_restore=ids_restore,
            )
            for domain in self.output_adapters
            if domain not in self.fp32_output_adapters
        }
        # Force running selected output adapters in fp32 mode
        with torch.amp.autocast('cuda', enabled=False):
            for domain in self.fp32_output_adapters:
                if domain not in self.output_adapters:
                    continue
                preds[domain] = self.output_adapters[domain](
                    encoder_tokens=encoder_tokens.float(),
                    input_info=input_info,
                    ids_keep=ids_keep,
                    ids_restore=ids_restore,
                )

        loss = {f'{domain}_loss': self.loss_functions[domain](pred, x[domain], task_masks[domain])
                for domain, pred in preds.items()}

        loss['loss'] = torch.stack(list(loss.values())).sum()

        # Convert token masks to pixel masks
        for key, mask in task_masks.items():
            # TODO: Assumes squared inputs
            N_sqrt = int(mask.shape[1] ** 0.5)
            mask = mask.view(B, N_sqrt, N_sqrt)
            task_masks[key] = F.interpolate(mask.unsqueeze(1).to(torch.uint8), size=(H, W), mode='nearest').squeeze(1)

        return loss, preds, task_masks


class MultiViT(MultiMAE):
    """MultiViT: Multi-modal Vision Transformer
    This is MultiMAE without masking and with a simplified / faster forward pass


    :param input_adapters: Dictionary of task -> input adapters
    :param output_adapters: Optional dictionary of task -> output adapters

    :param num_global_tokens: Number of additional global tokens to add (like cls tokens), default is 1
    :param dim_tokens: Dimension of encoder tokens
    :param depth: Depth of encoder
    :param num_heads: Number of attention heads
    :param mlp_ratio: MLP hidden dim ratio
    :param qkv_bias: Set to False to disable bias
    :param drop_rate: Dropout after MLPs and Attention
    :param attn_drop_rate: Attention matrix drop rate
    :param drop_path_rate: DropPath drop rate
    :param norm_layer: Type of normalization layer
    """

    def process_input(self, x):
        # If input x is a Tensor, assume it's RGB
        x = {"rgb": x} if isinstance(x, torch.Tensor) else x
        # Need image size for tokens->image reconstruction
        if "rgb" in x:
            B, _, H, W = x["rgb"].shape
        elif "semseg" in x:
            B, H, W = x["semseg"].shape
            H *= self.input_adapters["semseg"].stride_level
            W *= self.input_adapters["semseg"].stride_level
        else:
            # Assuming same shape for all inputs
            B, _, H, W = list(x.values())[0].shape

        # Encode selected inputs to tokens
        input_task_tokens = {
            domain: self.input_adapters[domain](tensor)
            for domain, tensor in x.items()
            if domain in self.input_adapters
        }

        input_info = self.generate_input_info(input_task_tokens=input_task_tokens, image_size=(H, W))
        input_tokens = torch.cat(list(input_task_tokens.values()), dim=1)

        # Add global tokens to input tokens
        global_tokens = repeat(self.global_tokens, "() n d -> b n d", b=B)
        input_tokens = torch.cat([input_tokens, global_tokens], dim=1)
        return input_tokens, input_info

    def process_output(self, x, input_info, method):
        # TODO: Make it generalizable
        num_tokens = list(input_info['tasks'].values())[0]['num_tokens']
        assert any(info['num_tokens'] == num_tokens for info in list(input_info['tasks'].values())), \
            "Current code only supports modalities with the same number of tokens"
        def _unstack_image_modalities(x):
            x = torch.split(x, num_tokens, dim=1)  # Split tokens by modality
            x = torch.stack(x, dim=1)  # (B, M, N, D)
            return x

        # Merge tokens from different modalities
        if self.merge_method == 'mean':
            x = _unstack_image_modalities(x)
            x = x.mean(dim=1)

        elif self.merge_method == 'max':
            x = _unstack_image_modalities(x)
            x = x.max(dim=1)[0]

        elif self.merge_method == 'concat':
            # TODO: Handle missing modalities with a learnable self.missing_token. Currently expects all modalities.
            assert len(input_info['tasks']) == len(self.input_adapters), "Method concat expects all modalities as input"
            x = _unstack_image_modalities(x)
            x = torch.cat(x.unbind(dim=1), dim=-1)

        elif self.merge_method is None:
            pass  # Do nothing
        else:
            raise NotImplementedError(f'Merging method {self.merge_method} is not implemented. '
                                      f'Select one of mean, max or concat.')

        return x


    def forward(
        self,
        x: dict[str, torch.Tensor] | torch.Tensor,
    ):
        """
        Forward pass through input adapters, transformer encoder and output adapters.

        :param x: Input tensor or dictionary of tensors
        :param return_all_layers: Set to True to return all transformer layers
        """

        x, input_info = self.process_input(x)

        out = []
        for block in self.layers:
            x = block(x)
            out.append(x)

        # Drop global token
        global_token = [tokens[:, :input_info['num_global_tokens']] for tokens in out]
        out = [tokens[:, input_info['num_global_tokens']:] for tokens in out]

        out = [self.process_output(x, input_info, self.merge_method) for x in out]

        return out



# Copyright contributors to the Terratorch project

from torch import Tensor, nn

from terratorch.registry import TERRATORCH_DECODER_REGISTRY


@TERRATORCH_DECODER_REGISTRY.register
class LinearDecoder(nn.Module):
    """
    A linear decoder using a transposed convolution layer for upsampling.
    """
    includes_head: bool = True

    def __init__(self, embed_dim: list[int], num_classes: int, upsampling_size: int, in_index: int = -1) -> None:
        """Constructor

        Args:
            embed_dim (list[int]): A list of embedding dimensions for different feature maps.
            num_classes (int): Number of output classes.
            upsampling_size (int): Kernel and stride size for transposed convolution.
            in_index (int, optional): Index of the input feature map to use. Defaults to -1."
        """
        super().__init__()
        self.num_classes = num_classes
        self.in_index = in_index
        self.embed_dim = embed_dim[in_index]

        self.conv = nn.ConvTranspose2d(
            in_channels=self.embed_dim,
            out_channels=self.num_classes,
            kernel_size=upsampling_size,
            stride=upsampling_size,
            padding=0,
            output_padding=0,
        )

    @property
    def out_channels(self) -> int:
        return self.num_classes

    def forward(self, x: list[Tensor]) -> Tensor:
        return self.conv(x[self.in_index])



import importlib
import inspect
import logging
from collections.abc import Callable, Set
import warnings

from torch import nn

from terratorch.registry import DECODER_REGISTRY


class MMsegDecoderWrapper(nn.Module):
    """
    A wrapper for decoders designed to handle single or multiple embeddings with specified indices.

    Attributes:
        decoder (nn.Module): The decoder module being wrapped.
        channels (int): The number of output channels of the decoder.
    Methods:
        forward(x: List[torch.Tensor]) -> torch.Tensor:
            Forward pass for embeddings with specified indices.
    """

    includes_head: bool = True

    def __init__(self, decoder) -> None:
        """
        Args:
            decoder (nn.Module): The decoder module to be wrapped.
            out_channels (int): Output channels of the decoder, needed for terratorch head.
                Necessary since not all smp decoders have the `out_channels` attribute.
            Defaults to -1.
        """
        super().__init__()
        self.decoder = decoder

    def forward(self, x):
        return self.decoder(x)


class MMSegRegistry(Set):
    """Registry wrapper for mmseg"""

    includes_head: bool = True

    def __init__(self):
        if not importlib.util.find_spec("mmseg"):
            msg = "mmsegmentation must be installed to instantiate an MMSegRegistry"
            raise ImportError(msg)
        self.mmseg_reg = importlib.import_module("mmseg.models.decode_heads")
        self.mmseg_decoder_names = {x for x, _ in inspect.getmembers(self.mmseg_reg, inspect.isclass)}

    def register(self, constructor: Callable | type) -> Callable:
        raise NotImplementedError()

    def build(self, name: str, in_channels, *constructor_args, **constructor_kwargs) -> nn.Module:
        """Build and return the component.
        Use prefixes ending with _ to forward to a specific source
        """
        try:
            decoder = getattr(self.mmseg_reg, name)
        except AttributeError as e:
            msg = f"Decoder {name} not found"
            raise KeyError(msg) from e

        if len(in_channels) == 1:
            in_channels = in_channels[0]
        if "num_classes" not in constructor_kwargs:
            msg = "num_classes is a required argument for mmseg decoders. If you are using an mmseg decoder for a regression task please include num_classes=1."
            raise ValueError(msg)
        
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning) # dont worry about warnings related to loss. we dont use that component.
            decoder = decoder(*constructor_args, in_channels=in_channels, **constructor_kwargs)
        return MMsegDecoderWrapper(decoder)

    def __iter__(self):
        return iter(self.mmseg_decoder_names)

    def __len__(self):
        return len(self.mmseg_decoder_names)

    def __contains__(self, key):
        return key in self.mmseg_decoder_names

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __str__(self):
        return f"Mmseg registry with {len(self)} registered backbones"

    def __getitem__(self, name):
        try:
            return getattr(self.mmseg_reg, name)
        except AttributeError as e:
            msg = f"Decoder {name} not found"
            raise KeyError(msg) from e


if importlib.util.find_spec("mmseg"):
    MMSEG_DECODER_REGISTRY = MMSegRegistry()
    DECODER_REGISTRY.register_source("mmseg", MMSEG_DECODER_REGISTRY)
else:
    logging.getLogger("terratorch").debug("mmseg not installed, so MmsegDecoderRegistry not created")



# Copyright contributors to the Terratorch project

"""Command-line interface to TerraTorch."""

from terratorch.cli_tools import build_lightning_cli
import sys
from huggingface_hub import hf_hub_download
import logging

try:
    from benchmark.main import main as iterate_main

    TERRATORCH_ITERATE_INSTALLED = True

except ImportError:
    TERRATORCH_ITERATE_INSTALLED = False


def main():
    if len(sys.argv) == 1:
        print('usage: terratorch [-h] [-c CONFIG] [--print_config[=flags]] {fit,validate,test,predict,compute_statistics,init,iterate} ...')
        exit(0)
    if len(sys.argv) >= 2 and sys.argv[1] == "iterate":
        # if user runs "terratorch iterate" and terratorch-iterate has not been installed
        if not TERRATORCH_ITERATE_INSTALLED:
            print(
                (
                    "Error! terratorch-iterate has not been installed. If you want to install it,"
                    "run 'pip install terratorch-iterate'"
                )
            )
        # if user runs "terratorch iterate" and terratorch-iterate has been installed
        else:
            # delete iterate argument
            del sys.argv[1]
            iterate_main()
    elif sys.argv[1] == "init":
        logger = logging.getLogger("terratorch-init")
        logger.info("Initializing TerraTorch...")
        logger.info("Downloading model congfigs...")
        files_to_download = [
            {
                'repo_id': 'ibm-granite/granite-geospatial-biomass',
            },
            {
                'repo_id': 'ibm-granite/granite-geospatial-wxc-downscaling',
            },
            {
                'repo_id': 'ibm-granite/granite-geospatial-canopyheight',
            },
            {
                'repo_id': 'ibm-granite/granite-geospatial-land-surface-temperature',
            },
            {
                'repo_id': 'ibm-granite/granite-geospatial-uki',
            },
            {
                'repo_id': 'ibm-granite/granite-geospatial-uki-flooddetection',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M-rollout',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-WxC-1.0-2300m-gravity-wave-parameterization',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-EO-2.0-300M-TL',
                'filename': 'config.json',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-EO-2.0-600M-TL',
                'filename': 'config.json',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-EO-2.0-300M-BurnScars',
                'filename': 'burn_scars_config.yaml',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-EO-1.0-100M-burn-scar',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-EO-1.0-100M-multi-temporal-crop-classification',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-EO-2.0-300M',
                'filename': 'config.json',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-EO-1.0-100M',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-EO-2.0-300M-TL-Sen1Floods11',
            },
            {
                'repo_id': 'ibm-nasa-geospatial/Prithvi-EO-1.0-100M-sen1floods11',
            },
        ]

        for config in files_to_download:
            file_path = hf_hub_download(
                repo_id=config['repo_id'],
                filename=config.get('filename', '') or 'config.yaml',
            )
            logger.info(f"File downloaded to: {file_path}")
    else:
        _ = build_lightning_cli()


if __name__ == "__main__":
    main()



# Copyright contributors to the Terratorch project

"""
This is just an example of a possible structure to include SMP models
Right now it always returns a UNET, but could easily be extended to many of the models provided by SMP.
"""

import importlib

import torch
from torch import nn

from terratorch.models.model import Model, ModelFactory, ModelOutput
from terratorch.models.utils import extract_prefix_keys
from terratorch.registry import MODEL_FACTORY_REGISTRY
from terratorch.tasks.segmentation_tasks import to_segmentation_prediction


def freeze_module(module: nn.Module):
    for param in module.parameters():
        param.requires_grad_(False)

@MODEL_FACTORY_REGISTRY.register
class GenericUnetModelFactory(ModelFactory):
    def _check_model_availability(self, model, builtin_engine, engine, **model_kwargs):

        try:
            print(f"Using module {model} from terratorch.")
            if builtin_engine:
                model_class = getattr(builtin_engine, model)
            else:
                model_class = None
        except:
            if _has_mmseg:
                print("Module not available on terratorch.")
                print(f"Using module {model} from mmseg.")
                if engine:
                    model_class = getattr(engine, model)
                else:
                    model_class = None
            else:
                raise Exception("mmseg is not installed.")

        if model_class:
            model = model_class(
               **model_kwargs,
            )
        else:
            model = None

        return model 

    def build_model(
        self,
        task: str = "segmentation",
        backbone: str | None = None,
        decoder: str | None = None,
        dilations: tuple[int] = (1, 6, 12, 18),
        in_channels: int = 6,
        pretrained: str | bool | None = True,
        regression_relu: bool = False,
        **kwargs,
    ) -> Model:
        """Factory to create model based on mmseg.

        Args:
            task (str): Must be "segmentation".
            model (str): Decoder architecture. Currently only supports "unet".
            in_channels (int): Number of input channels.
            pretrained(str | bool): Which weights to use for the backbone. If true, will use "imagenet". If false or None, random weights. Defaults to True.
            regression_relu (bool). Whether to apply a ReLU if task is regression. Defaults to False.

        Returns:
            Model: UNet model.
        """
        if task not in ["segmentation", "regression"]:
            msg = f"This model can only perform pixel wise tasks, but got task {task}"
            raise Exception(msg)

        builtin_engine_decoders = importlib.import_module("terratorch.models.decoders")
        builtin_engine_encoders = importlib.import_module("terratorch.models.backbones")

        # Default values
        backbone_builtin_engine = None
        decoder_builtin_engine = None
        backbone_engine = None 
        decoder_engine = None 
        backbone_model_kwargs = {}
        decoder_model_kwargs = {}

        try:
            engine_decoders = importlib.import_module("mmseg.models.decode_heads")
            engine_encoders = importlib.import_module("mmseg.models.backbones")
            _has_mmseg = True
        except:
            engine_decoders = None
            engine_encoders = None
            _has_mmseg = False
            print("mmseg is not installed.")

        if backbone:
            backbone_kwargs = _extract_prefix_keys(kwargs, "backbone_")
            backbone_model_kwargs = backbone_kwargs
            backbone_engine = engine_encoders
            backbone_builtin_engine = builtin_engine_encoders
        else:
            backbone=None

        if decoder: 
            decoder_kwargs = _extract_prefix_keys(kwargs, "decoder_")
            decoder_model_kwargs = decoder_kwargs
            decoder_engine = engine_decoders
            decoder_builtin_engine = builtin_engine_decoders
        else:
            decoder = None 

        if not backbone and not decoder:
            print("It is necessary to define a backbone and/or a decoder.")

        # Instantianting backbone and decoder 
        backbone = self._check_model_availability(backbone, backbone_builtin_engine, backbone_engine, **backbone_model_kwargs) 
        decoder = self._check_model_availability(decoder, decoder_builtin_engine, decoder_engine, **decoder_model_kwargs) 

        return GenericUnetModelWrapper(
            backbone, decoder=decoder, relu=task == "regression" and regression_relu, squeeze_single_class=task == "regression"
        )

class GenericUnetModelWrapper(Model, nn.Module):
    def __init__(self, unet_model, decoder=None, relu=False, squeeze_single_class=False, intermediary_outputs_in_decoder=False) -> None:
        super().__init__()
        self.unet_model = unet_model
        self.decoder = decoder
        self.final_act = nn.ReLU() if relu else nn.Identity()
        self.squeeze_single_class = squeeze_single_class
        self.intermediary_outputs_in_decoder = intermediary_outputs_in_decoder

        if decoder:
            self.decode = self._with_decoder
        else:
            self.decode = self._no_decoder

        if self.intermediary_outputs_in_decoder:
            self.catch_unet_outputs = lambda x: x
        else:
            self.catch_unet_outputs =  lambda x: x[-1]

    def _no_decoder(self, x):
        return x

    def _with_decoder(self, x):

        return self.decoder(x)

    def forward(self, *args, **kwargs):

        # It supposes the input has dimension (B, C, H, W)
        input_data = [args[0]] # It adapts the input to became a list of time 'snapshots'
        args = input_data

        unet_output = self.unet_model(*args, **kwargs)
        unet_output = self.final_act(unet_output)

        unet_output = self.catch_unet_outputs(unet_output)
        # Decoding is optional
        unet_output_decoded = self.decode(unet_output)

        if unet_output_decoded.shape[1] == 1 and self.squeeze_single_class:
            unet_output_decoded = unet_output_decoded.squeeze(1)

        model_output = ModelOutput(unet_output_decoded)

        return model_output

    def freeze_encoder(self):
        raise freeze_module(self.unet_model)

    def freeze_decoder(self):
        raise freeze_module(self.decoder)


def _extract_prefix_keys(d: dict, prefix: str) -> dict:
    extracted_dict = {}
    keys_to_del = []
    for k, v in d.items():
        if k.startswith(prefix):
            extracted_dict[k.split(prefix)[1]] = v
            keys_to_del.append(k)

    for k in keys_to_del:
        del d[k]

    return extracted_dict




from typing import Any
import torch

import albumentations as A
import kornia.augmentation as K
from torchgeo.datamodules import NonGeoDataModule
from kornia.augmentation import AugmentationSequential
from terratorch.datasets import OpenEarthMapNonGeo
from terratorch.datamodules.utils import wrap_in_compose_is_list

MEANS = {
    "BLUE": 116.628328,
    "GREEN": 119.65935,
    "RED": 113.385309
}

STDS = {
    "BLUE": 44.668890717415586,
    "GREEN": 48.282311849967364,
    "RED": 54.19692448815262,
}

class OpenEarthMapNonGeoDataModule(NonGeoDataModule):
    """NonGeo LightningDataModule implementation for Open Earth Map."""

    def __init__(
        self, 
        batch_size: int = 8, 
        num_workers: int = 0, 
        data_root: str = "./",
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        predict_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        **kwargs: Any
    ) -> None:
        """
        Initializes the OpenEarthMapNonGeoDataModule for the Open Earth Map dataset.

        Args:
            batch_size (int, optional): Batch size for DataLoaders. Defaults to 8.
            num_workers (int, optional): Number of workers for data loading. Defaults to 0.
            data_root (str, optional): Root directory of the dataset. Defaults to "./".
            train_transform (A.Compose | None | list[A.BasicTransform], optional): Transformations for training data.
            val_transform (A.Compose | None | list[A.BasicTransform], optional): Transformations for validation data.
            test_transform (A.Compose | None | list[A.BasicTransform], optional): Transformations for test data.
            predict_transform (A.Compose | None | list[A.BasicTransform], optional): Transformations for prediction data.
            aug (AugmentationSequential, optional): Augmentation pipeline; if None, defaults to normalization using computed means and stds.
            **kwargs: Additional keyword arguments. Can include 'bands' (list[str]) to specify the bands; defaults to OpenEarthMapNonGeo.all_band_names if not provided.
        """
        super().__init__(OpenEarthMapNonGeo, batch_size, num_workers, **kwargs)

        bands = kwargs.get("bands", OpenEarthMapNonGeo.all_band_names)
        self.means = torch.tensor([MEANS[b] for b in bands])
        self.stds = torch.tensor([STDS[b] for b in bands])
        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)
        self.predict_transform = wrap_in_compose_is_list(predict_transform)
        self.data_root = data_root
        self.aug = AugmentationSequential(K.Normalize(self.means, self.stds), data_keys=None) if aug is None else aug

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either fit, validate, test, or predict.
        """
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(  
                split="train", data_root=self.data_root, transform=self.train_transform, **self.kwargs
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                split="val", data_root=self.data_root, transform=self.val_transform, **self.kwargs
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                split="test",data_root=self.data_root, transform=self.test_transform, **self.kwargs
            )
        if stage in ["predict"]:
            self.predict_dataset = self.dataset_class(
                split="test",data_root=self.data_root, transform=self.predict_transform, **self.kwargs
            )



# ------------------------------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------------------------------
# Modified from https://github.com/chengdazhi/Deformable-Convolution-V2-PyTorch/tree/pytorch_1.0.0
# ------------------------------------------------------------------------------------------------


import math
import warnings
from typing import Any

import torch
import torch.nn.functional as F  # noqa: N812
from torch import nn
from torch.nn.init import constant_, xavier_uniform_

from terratorch.models.backbones.detr_ops.functions import MSDeformAttnFunction


def _is_power_of_2(n: Any) -> bool:
    if (not isinstance(n, int)) or (n < 0):
        msg = f"invalid input for _is_power_of_2: {n} (type: {type(n)})"
        raise ValueError(msg)
    return (n & (n - 1) == 0) and n != 0


class MSDeformAttn(nn.Module):
    def __init__(
        self,
        d_model: int = 256,
        n_levels: int = 4,
        n_heads: int = 8,
        n_points: int = 4,
        ratio: float = 1.0,
    ):
        """Multi-Scale Deformable Attention Module.

        Args:
            d_model (int): Hidden dimension.
            n_levels (int): Number of feature levels.
            n_heads (int): Number of attention heads.
            n_points (int): Number of sampling points per attention head per feature level.
        """
        super().__init__()
        if d_model % n_heads != 0:
            msg = f"d_model must be divisible by n_heads, but got {d_model} and {n_heads}"
            raise ValueError(msg)
        _d_per_head = d_model // n_heads
        # you'd better set _d_per_head to a power of 2
        # which is more efficient in our CUDA implementation
        if not _is_power_of_2(_d_per_head):
            warnings.warn(
                "You'd better set d_model in MSDeformAttn to make "
                "the dimension of each attention head a power of 2 "
                "which is more efficient in our CUDA implementation.",
                stacklevel=1,
            )

        self.im2col_step = 64

        self.d_model = d_model
        self.n_levels = n_levels
        self.n_heads = n_heads
        self.n_points = n_points
        self.ratio = ratio
        self.sampling_offsets = nn.Linear(d_model, n_heads * n_levels * n_points * 2)
        self.attention_weights = nn.Linear(d_model, n_heads * n_levels * n_points)
        self.value_proj = nn.Linear(d_model, int(d_model * ratio))
        self.output_proj = nn.Linear(int(d_model * ratio), d_model)

        self._reset_parameters()

    def _reset_parameters(self) -> None:
        constant_(self.sampling_offsets.weight.data, 0.0)
        thetas = torch.arange(self.n_heads, dtype=torch.float32) * (2.0 * math.pi / self.n_heads)
        grid_init = torch.stack([thetas.cos(), thetas.sin()], -1)
        grid_init = (
            (grid_init / grid_init.abs().max(-1, keepdim=True)[0])
            .view(self.n_heads, 1, 1, 2)
            .repeat(1, self.n_levels, self.n_points, 1)
        )
        for i in range(self.n_points):
            grid_init[:, :, i, :] *= i + 1

        with torch.no_grad():
            self.sampling_offsets.bias = nn.Parameter(grid_init.view(-1))
        constant_(self.attention_weights.weight.data, 0.0)
        constant_(self.attention_weights.bias.data, 0.0)
        xavier_uniform_(self.value_proj.weight.data)
        constant_(self.value_proj.bias.data, 0.0)
        xavier_uniform_(self.output_proj.weight.data)
        constant_(self.output_proj.bias.data, 0.0)

    def forward(
        self,
        query: torch.Tensor,
        reference_points: torch.Tensor,
        input_flatten: torch.Tensor,
        input_spatial_shapes: torch.Tensor,
        input_level_start_index: torch.Tensor,
        input_padding_mask: torch.Tensor | None = None,
    ) -> torch.Tensor:
        r"""
        Args:
            query (torch.Tensor): (N, Length_query, C)
            reference_points (torch.Tensor): (N, Length_{query}, n_levels, 2), range in [0, 1], top-left (0,0),
                bottom-right (1, 1), including padding area or (N, Length_{query}, n_levels, 4), add additional (w, h)
                to form reference boxes.
            input_flatten (torch.Tensor): (N, \sum_{l=0}^{L-1} H_l \cdot W_l, C)
            input_spatial_shapes (torch.Tensor): (n_levels, 2), [(H_0, W_0), (H_1, W_1), ..., (H_{L-1}, W_{L-1})]
            input_level_start_index (torch.Tensor): (n_levels, ), [0, H_0*W_0, H_0*W_0+H_1*W_1, H_0*W_0+H_1*W_1+H_2*W_2,
                ..., H_0*W_0+H_1*W_1+...+H_{L-1}*W_{L-1}]
            input_padding_mask (torch.Tensor): (N, \sum_{l=0}^{L-1} H_l \cdot W_l), True for padding elements, False for
                non-padding elements.
        """

        N, Len_q, _ = query.shape  # noqa: N806
        N, Len_in, _ = input_flatten.shape  # noqa: N806
        assert (input_spatial_shapes[:, 0] * input_spatial_shapes[:, 1]).sum() == Len_in

        value = self.value_proj(input_flatten)
        if input_padding_mask is not None:
            value = value.masked_fill(input_padding_mask[..., None], float(0))

        value = value.view(N, Len_in, self.n_heads, int(self.ratio * self.d_model) // self.n_heads)
        sampling_offsets = self.sampling_offsets(query).view(N, Len_q, self.n_heads, self.n_levels, self.n_points, 2)
        attention_weights = self.attention_weights(query).view(N, Len_q, self.n_heads, self.n_levels * self.n_points)
        attention_weights = F.softmax(attention_weights, -1).view(N, Len_q, self.n_heads, self.n_levels, self.n_points)

        if reference_points.shape[-1] == 2:
            offset_normalizer = torch.stack([input_spatial_shapes[..., 1], input_spatial_shapes[..., 0]], -1)
            sampling_locations = (
                reference_points[:, :, None, :, None, :]
                + sampling_offsets / offset_normalizer[None, None, None, :, None, :]
            )
        elif reference_points.shape[-1] == 4:
            sampling_locations = (
                reference_points[:, :, None, :, None, :2]
                + sampling_offsets / self.n_points * reference_points[:, :, None, :, None, 2:] * 0.5
            )
        else:
            msg = f"Last dim of reference_points must be 2 or 4, but get {reference_points.shape[-1]} instead."
            raise ValueError(msg)
        output: torch.Tensor = MSDeformAttnFunction.apply(  # type: ignore
            value,
            input_spatial_shapes,
            input_level_start_index,
            sampling_locations,
            attention_weights,
            self.im2col_step,
        )
        output = self.output_proj(output)
        return output



from abc import ABC, abstractmethod

import numpy as np
import torch
import torch.nn.functional as F
from einops import rearrange
from torch import nn
import math

from terratorch.registry import NECK_REGISTRY, TERRATORCH_NECK_REGISTRY


class Neck(ABC, nn.Module):
    """Base class for Neck

    A neck must must implement `self.process_channel_list` which returns the new channel list.
    """

    def __init__(self, channel_list: list[int]) -> None:
        super().__init__()
        self.channel_list = channel_list

    @abstractmethod
    def process_channel_list(self, channel_list: list[int]) -> list[int]:
        return channel_list

    @abstractmethod
    def forward(self, channel_list: list[torch.Tensor]) -> list[torch.Tensor]: ...


@TERRATORCH_NECK_REGISTRY.register
class SelectIndices(Neck):
    def __init__(self, channel_list: list[int], indices: list[int]):
        """Select indices from the embedding list

        Args:
            indices (list[int]): list of indices to select.
        """
        super().__init__(channel_list)
        self.indices = indices

    def forward(self, features: list[torch.Tensor]) -> list[torch.Tensor]:
        features = [features[i] for i in self.indices]
        return features

    def process_channel_list(self, channel_list: list[int]) -> list[int]:
        channel_list = [channel_list[i] for i in self.indices]
        return channel_list

@TERRATORCH_NECK_REGISTRY.register
class PermuteDims(Neck):
    def __init__(self, channel_list: list[int], new_order: list[int]):
        """Permute dimensions of each element in the embedding list

        Args:
            new_order (list[int]): list of indices to be passed to tensor.permute()
        """
        super().__init__(channel_list)
        self.new_order = new_order

    def forward(self, features: list[torch.Tensor]) -> list[torch.Tensor]:
        features = [feat.permute(*self.new_order).contiguous() for feat in features]
        return features

    def process_channel_list(self, channel_list: list[int]) -> list[int]:
        return super().process_channel_list(channel_list)

@TERRATORCH_NECK_REGISTRY.register
class InterpolateToPyramidal(Neck):
    def __init__(self, channel_list: list[int], scale_factor: int = 2, mode: str = "nearest"):
        """Spatially interpolate embeddings so that embedding[i - 1] is scale_factor times larger than embedding[i]

        Useful to make non-pyramidal backbones compatible with hierarachical ones
        Args:
            scale_factor (int): Amount to scale embeddings by each layer. Defaults to 2.
            mode (str): Interpolation mode to be passed to torch.nn.functional.interpolate. Defaults to 'nearest'.
        """
        super().__init__(channel_list)
        self.scale_factor = scale_factor
        self.mode = mode

    def forward(self, features: list[torch.Tensor]) -> list[torch.Tensor]:
        out = []
        scale_exponents = list(range(len(features), 0, -1))
        for x, exponent in zip(features, scale_exponents, strict=True):
            out.append(F.interpolate(x, scale_factor=self.scale_factor**exponent, mode=self.mode))

        return out

    def process_channel_list(self, channel_list: list[int]) -> list[int]:
        return super().process_channel_list(channel_list)


@TERRATORCH_NECK_REGISTRY.register
class MaxpoolToPyramidal(Neck):
    def __init__(self, channel_list: list[int], kernel_size: int = 2):
        """Spatially downsample embeddings so that embedding[i - 1] is scale_factor times smaller than embedding[i]

        Useful to make non-pyramidal backbones compatible with hierarachical ones
        Args:
            kernel_size (int). Base kernel size to use for maxpool. Defaults to 2.
        """
        super().__init__(channel_list)
        self.kernel_size = kernel_size

    def forward(self, features: list[torch.Tensor]) -> list[torch.Tensor]:
        out = []
        scale_exponents = list(range(len(features)))
        for x, exponent in zip(features, scale_exponents, strict=True):
            if exponent == 0:
                out.append(x.clone())
            else:
                out.append(F.max_pool2d(x, kernel_size=self.kernel_size**exponent))

        return out

    def process_channel_list(self, channel_list: list[int]) -> list[int]:
        return super().process_channel_list(channel_list)


@TERRATORCH_NECK_REGISTRY.register
class ReshapeTokensToImage(Neck):
    def __init__(self, channel_list: list[int], remove_cls_token=True, effective_time_dim: int = 1):  # noqa: FBT002
        """Reshape output of transformer encoder so it can be passed to a conv net.

        Args:
            remove_cls_token (bool, optional): Whether to remove the cls token from the first position.
                Defaults to True.
            effective_time_dim (int, optional): The effective temporal dimension the transformer processes.
                For a ViT, his will be given by `num_frames // tubelet size`. This is used to determine
                the temporal dimension of the embedding, which is concatenated with the embedding dimension.
                For example:
                - A model which processes 1 frame with a tubelet size of 1 has an effective_time_dim of 1.
                    The embedding produced by this model has embedding size embed_dim * 1.
                - A model which processes 3 frames with a tubelet size of 1 has an effective_time_dim of 3.
                    The embedding produced by this model has embedding size embed_dim * 3.
                - A model which processes 12 frames with a tubelet size of 4 has an effective_time_dim of 3.
                    The embedding produced by this model has an embedding size embed_dim * 3.
                Defaults to 1.
        """
        super().__init__(channel_list)
        self.remove_cls_token = remove_cls_token
        self.effective_time_dim = effective_time_dim

    def collapse_dims(self, x):
        """
        When the encoder output has more than 3 dimensions, is necessary to 
        reshape it. 
        """
        shape = x.shape
        batch = x.shape[0]
        e = x.shape[-1]
        collapsed_dim = np.prod(x.shape[1:-1])

        return x.reshape(batch, collapsed_dim, e)

    def forward(self, features: list[torch.Tensor]) -> list[torch.Tensor]:
        out = []
        for x in features:
            if self.remove_cls_token:
                x_no_token = x[:, 1:, :]
            else:
                x_no_token = x
            x_no_token = self.collapse_dims(x_no_token)
            number_of_tokens = x_no_token.shape[1]
            tokens_per_timestep = number_of_tokens // self.effective_time_dim
            h = int(math.sqrt(tokens_per_timestep))

            encoded = rearrange(
                x_no_token,
                "batch (t h w) e -> batch (t e) h w",
                batch=x_no_token.shape[0],
                t=self.effective_time_dim,
                h=h,
            )
            out.append(encoded)
        return out

    def process_channel_list(self, channel_list: list[int]) -> list[int]:
        return super().process_channel_list(channel_list)

@TERRATORCH_NECK_REGISTRY.register
class AddBottleneckLayer(Neck):
    """Add a layer that reduces the channel dimension of the final embedding by half, and concatenates it

    Useful for compatibility with some smp decoders.
    """

    def __init__(self, channel_list: list[int]):
        super().__init__(channel_list)
        self.bottleneck = nn.Conv2d(channel_list[-1], channel_list[-1]//2, kernel_size=1)

    def forward(self, features: list[torch.Tensor]) -> list[torch.Tensor]:
        new_embedding = self.bottleneck(features[-1])
        features.append(new_embedding)
        return features

    def process_channel_list(self, channel_list: list[int]) -> list[int]:
        return [*channel_list, channel_list[-1] // 2]

@TERRATORCH_NECK_REGISTRY.register
class LearnedInterpolateToPyramidal(Neck):
    """Use learned convolutions to transform the output of a non-pyramidal encoder into pyramidal ones

    Always requires exactly 4 embeddings
    """

    def __init__(self, channel_list: list[int]):
        super().__init__(channel_list)
        if len(channel_list) != 4:
            msg = "This class can only handle exactly 4 input embeddings"
            raise Exception(msg)
        self.fpn1 = nn.Sequential(
            nn.ConvTranspose2d(channel_list[0], channel_list[0] // 2, 2, 2),
            nn.BatchNorm2d(channel_list[0] // 2),
            nn.GELU(),
            nn.ConvTranspose2d(channel_list[0] // 2, channel_list[0] // 4, 2, 2),
        )
        self.fpn2 = nn.Sequential(nn.ConvTranspose2d(channel_list[1], channel_list[1] // 2, 2, 2))
        self.fpn3 = nn.Sequential(nn.Identity())
        self.fpn4 = nn.Sequential(nn.MaxPool2d(kernel_size=2, stride=2))
        self.embedding_dim = [channel_list[0] // 4, channel_list[1] // 2, channel_list[2], channel_list[3]]

    def forward(self, features: list[torch.Tensor]) -> list[torch.Tensor]:
        scaled_inputs = []
        scaled_inputs.append(self.fpn1(features[0]))
        scaled_inputs.append(self.fpn2(features[1]))
        scaled_inputs.append(self.fpn3(features[2]))
        scaled_inputs.append(self.fpn4(features[3]))
        return scaled_inputs

    def process_channel_list(self, channel_list: list[int]) -> list[int]:
        return [channel_list[0] // 4, channel_list[1] // 2, channel_list[2], channel_list[3]]


def build_neck_list(ops: list[dict], channel_list: list[int]) -> tuple[list[Neck], list[int]]:
    neck_list = []
    cur_channel_list = channel_list.copy()
    for cur_op in ops:
        op: Neck = NECK_REGISTRY.build(
            cur_op["name"], cur_channel_list, **{k: v for k, v in cur_op.items() if k != "name"}
        )
        cur_channel_list = op.process_channel_list(cur_channel_list)
        neck_list.append(op)

    return neck_list, cur_channel_list
